"""Module for clustering analysis.

This module contains functions to perform clustering analysis on neuron
activations.
"""
import numpy as np
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import pdist, squareform


def create_correlation_clusters(
    X, use_abs_correlation=True, clustering_threshold=0.5, method="average"
):
    """
    Create clusters based on neuron activation correlation. All neurons in the
    same cluster have "highly correlated" neurons that fire similarly on similar
    inputs.

    Parameters
    ----------
    X : numpy.ndarray
        Matrix of size [ NUM_TOKENS x NUM_NEURONS]. Usually the output of
        interpretation.utils.create_tensors
    use_abs_correlation : bool, optional
        Whether to use absolute correlation values. Two neurons that are correlated
        in the opposite direction may represent the same "knowledge" in a large
        neural network.
    clustering_threshold : float, optional
        Hyperparameter for clustering. This is used as the threshold to convert
        hierarchical clusters into flat clusters.

    Returns
    -------
    cluster_labels : list
        List of cluster labels for every neuron

    """
    # Compute correlations
    corr = np.corrcoef(X.T)
    corr = np.nan_to_num(corr)
    corr = (corr + corr.T) / 2
    np.fill_diagonal(corr, 1)
    print("Correlation matrix size (#neurons x #neurons):", corr.shape)

    # Cluster based on correlations
    if use_abs_correlation:
        dissimilarity = 1 - np.abs(corr)
    else:
        dissimilarity = 1 - corr
    hierarchy = linkage(squareform(dissimilarity), method=method)
    labels = fcluster(hierarchy, clustering_threshold, criterion="distance")
    print("Number of clusters detected: %d" % np.max(labels))

    return labels


def extract_independent_neurons(X, use_abs_correlation=True, clustering_threshold=0.5):
    """
    Extract independent neurons from the given set of neurons.

    This method first clusters all of the given neurons with every cluster
    representing similar neurons. A single neuron is then picked randomly from
    every cluster and this forms the final set of independent neurons that is
    returned

    .. seealso::
        `Dalvi, Fahim, et al. "Analyzing redundancy in pretrained transformer models." Proceedings of the 2020 Conference on Empirical Methods in Natural Language Processing (EMNLP). 2020. <https://www.aclweb.org/anthology/2020.emnlp-main.398.pdf>`_

    Parameters
    ----------
    X : numpy.ndarray
        Matrix of size [ NUM_TOKENS x NUM_NEURONS]. Usually the output of
        interpretation.utils.create_tensors
    use_abs_correlation : bool, optional
        Whether to use absolute correlation values. Two neurons that are correlated
        in the opposite direction may represent the same "knowledge" in a large
        neural network.
    clustering_threshold : float, optional
        Hyperparameter for clustering. This is used as the threshold to convert
        hierarchical clusters into flat clusters.

    Returns
    -------
    independent_neurons : list
        List of non-redundant indepenent neurons

    """
    clusters = create_correlation_clusters(X, use_abs_correlation, clustering_threshold)

    independent_neurons = []

    for i in range(np.min(clusters), np.max(clusters) + 1):
        independent_neurons.append(np.random.choice(np.where(clusters == i)[0]))

    return independent_neurons, clusters


def print_clusters(cluster_labels):
    """
    Utility function for printing clusters

    Parameters
    ----------
    cluster_labels : list
        List of cluster labels for every neuron. Usually the output of
        ``interpretation.clustering.create_correlation_clusters``.

    """
    for i in range(np.min(cluster_labels), np.max(cluster_labels) + 1):
        print(
            "Cluster %05d: %s"
            % (i, " ".join([str(x) for x in np.where(cluster_labels == i)[0]]))
        )


def scikit_extract_independent_neurons(X, clustering_threshold=0.5):
    """
    Alternative implementation of ``interpretation.clustering.extract_independent_neurons``.

    This is an alternative implementation of the ``extract_independent_neurons``
    function using scikit-learn to create the correlation matrix instead of
    numpy. Should give identical results.

    Parameters
    ----------
    X : numpy.ndarray
        Matrix of size [ NUM_TOKENS x NUM_NEURONS]. Usually the output of
        interpretation.utils.create_tensors
    clustering_threshold : float, optional
        Hyperparameter for clustering. This is used as the threshold to convert
        hierarchical clusters into flat clusters.

    Returns
    -------
    independent_neurons : list
        List of non-redundant indepenent neurons
    cluster_labels : list
        List of cluster labels for every neuron

    """
    c = pdist(X.T, metric="correlation")
    hi = linkage(c, method="average")
    clusters = fcluster(hi, clustering_threshold, criterion="distance")

    independent_neurons = []
    for i in range(np.min(clusters), np.max(clusters) + 1):
        independent_neurons.append(np.random.choice(np.where(clusters == i)[0]))

    return independent_neurons, clusters



"""Utility functions to manage representations.

This module contains functions that will help in managing extracted
representations, specifically on sub-word based data.
"""
import numpy as np
from tqdm import tqdm


def bpe_get_avg_activations(tokens, activations):
    """Aggregates activations by averaging assuming BPE-based tokenization.

    Given loaded tokens data and activations, this function aggeregates
    activations based on tokenized text. BPE based tokenization is assumed,
    with every non-terminal subword ending with "@@". The activations are
    aggregated by averaging over subwords.

    .. warning::
        This function is deprecated and will be removed in future versions.

    Parameters
    ----------
    tokens : dict
        Dictionary containing three lists, ``source``, ``source_aux`` and
        ``target``. Usually the output of ``data.loader.load_aux_data``.
    activations : list of numpy.ndarray
        Activations returned from ``loader.load_activations``.

    Returns
    -------
    activations : list of numpy.ndarray
        Subword aggregated activations corresponding to one per actual token
        found in the untokenized text.

    """
    all_activations = []
    num_neurons = activations[0].size(1)

    for i in range(0, len(tokens["source_aux"])):
        sourceIndex = 0
        thisBPE = ""
        source = tokens["source"][i]
        source_aux = tokens["source_aux"][i]
        num_words = len(source)
        new_activations = np.zeros((num_words, num_neurons))

        word_boundaries = []

        for j in range(0, len(tokens["source_aux"][i])):
            currSourceWord = tokens["source"][i][sourceIndex]
            thisBPE = thisBPE + tokens["source_aux"][i][j]

            if thisBPE != currSourceWord:
                thisBPE = thisBPE[:-2]
            else:
                word_boundaries.append(j)
                sourceIndex = sourceIndex + 1
                thisBPE = ""

        assert len(word_boundaries) == num_words

        prev_idx = 0
        for word_idx, boundary in enumerate(word_boundaries):
            avg_vector = np.average(activations[i][prev_idx : boundary + 1, :], axis=0)
            new_activations[word_idx, :] = avg_vector
            prev_idx = boundary + 1

        all_activations.append(new_activations)

    return all_activations


def bpe_get_last_activations(tokens, activations, is_brnn=True):
    """Aggregates activations by picking the last subword assuming BPE-based tokenization.

    Given loaded tokens data and activations, this function aggeregates
    activations based on tokenized text. BPE based tokenization is assumed,
    with every non-terminal subword ending with "@@". The activations are
    aggregated by picking the last subword for any given word.

    .. warning::
        This function is deprecated and will be removed in future versions.

    Parameters
    ----------
    tokens : dict
        Dictionary containing three lists, ``source``, ``source_aux`` and
        ``target``. Usually the output of ``data.loader.load_aux_data``.
    activations : list of numpy.ndarray
        Activations returned from ``loader.load_activations``.
    is_brnn : bool, optional
        Whether the model from which activations were extracted was bidirectional.
        Only applies for RNN models.

    Returns
    -------
    activations : list of numpy.ndarray
        Subword aggregated activations corresponding to one per actual token
        found in the untokenized text.

    """
    all_activations = []
    num_neurons = activations[0].size(1)

    for i in range(0, len(tokens["source_aux"])):
        sourceIndex = 0
        thisBPE = ""
        source = tokens["source"][i]
        source_aux = tokens["source_aux"][i]
        num_words = len(source)
        new_activations = np.zeros((num_words, num_neurons))

        word_boundaries = []

        for j in range(0, len(tokens["source_aux"][i])):
            currSourceWord = tokens["source"][i][sourceIndex]
            thisBPE = thisBPE + tokens["source_aux"][i][j]

            if thisBPE != currSourceWord:
                thisBPE = thisBPE[:-2]
            else:
                word_boundaries.append(j)
                sourceIndex = sourceIndex + 1
                thisBPE = ""

        assert len(word_boundaries) == num_words

        rnn_boundary = int(num_neurons / 2)
        if not is_brnn:
            rnn_boundary = num_neurons

        prev_idx = 0
        for word_idx, boundary in enumerate(word_boundaries):
            # 0 - num_neurons/2: Forward
            # num_neurons/2 - : Backward
            new_activations[word_idx, :rnn_boundary] = activations[i][
                boundary, :rnn_boundary
            ]
            if is_brnn:
                new_activations[word_idx, rnn_boundary:] = activations[i][
                    prev_idx, rnn_boundary:
                ]
            prev_idx = boundary + 1

        all_activations.append(new_activations)

    return all_activations


def char_get_avg_activations(tokens, activations):
    """Aggregates activations by averaging assuming Character-based tokenization.

    Given loaded tokens data and activations, this function aggeregates
    activations based on character-tokenized text. The activations are
    aggregated by averaging over characters.

    .. warning::
        This function is deprecated and will be removed in future versions.

    Parameters
    ----------
    tokens : dict
        Dictionary containing three lists, ``source``, ``source_aux`` and
        ``target``. Usually the output of ``data.loader.load_aux_data``.
    activations : list of numpy.ndarray
        Activations returned from ``loader.load_activations``.

    Returns
    -------
    activations : list of numpy.ndarray
        Character aggregated activations corresponding to one per actual token
        found in the untokenized text.

    """
    all_activations = []
    num_neurons = activations[0].size(1)

    for i in tqdm(range(0, len(tokens["source_aux"]))):
        sourceIndex = 0
        thisChar = ""
        source = tokens["source"][i]
        source_aux = tokens["source_aux"][i]
        num_words = len(source)
        new_activations = np.zeros((num_words, num_neurons))

        word_boundaries = []

        for word_idx, word in enumerate(tokens["source"][i]):
            if word_idx == 0:
                word_boundaries.append(len(word) - 1)
            else:
                word_boundaries.append(len(word) + 1 + word_boundaries[-1])

        if len(word_boundaries) != num_words:
            print(i, len(word_boundaries), num_words)
        assert len(word_boundaries) == num_words
        assert (
            tokens["source_aux"][i].count("_") + 1 - tokens["source"][i].count("_")
            == num_words
        ), (
            "Number of words dont match! (line: %d, source: %d, aux: %d)\n%s\n%s"
            % (
                i + 1,
                num_words,
                tokens["source_aux"][i].count("_") + 1,
                " ".join(tokens["source"][i]),
                " ".join(tokens["source_aux"][i]),
            )
        )

        prev_idx = 0
        for word_idx, boundary in enumerate(word_boundaries):
            avg_vector = np.average(activations[i][prev_idx : boundary + 1, :], axis=0)
            new_activations[word_idx, :] = avg_vector
            prev_idx = boundary + 2

        all_activations.append(new_activations)

    return all_activations


def char_get_last_activations(tokens, activations, is_brnn=True):
    """Aggregates activations by picking the last subword assuming Character-based tokenization.

    Given loaded tokens data and activations, this function aggeregates
    activations based on character-tokenized text. The activations are
    aggregated by picking the last character for any given word.

    .. warning::
        This function is deprecated and will be removed in future versions.

    Parameters
    ----------
    tokens : dict
        Dictionary containing three lists, ``source``, ``source_aux`` and
        ``target``. Usually the output of ``data.loader.load_aux_data``.
    activations : list of numpy.ndarray
        Activations returned from ``loader.load_activations``.
    is_brnn : bool, optional
        Whether the model from which activations were extracted was bidirectional.
        Only applies for RNN models.

    Returns
    -------
    activations : list of numpy.ndarray
        Character aggregated activations corresponding to one per actual token
        found in the untokenized text.

    """
    all_activations = []
    num_neurons = activations[0].size(1)

    for i in tqdm(range(0, len(tokens["source_aux"]))):
        sourceIndex = 0
        thisChar = ""
        source = tokens["source"][i]
        source_aux = tokens["source_aux"][i]
        num_words = len(source)
        new_activations = np.zeros((num_words, num_neurons))

        word_boundaries = []

        for word_idx, word in enumerate(tokens["source"][i]):
            if word_idx == 0:
                word_boundaries.append(len(word) - 1)
            else:
                word_boundaries.append(len(word) + 1 + word_boundaries[-1])

        if len(word_boundaries) != num_words:
            print(i, len(word_boundaries), num_words)
        assert len(word_boundaries) == num_words
        assert (
            tokens["source_aux"][i].count("_") + 1 - tokens["source"][i].count("_")
            == num_words
        ), (
            "Number of words dont match! (line: %d, source: %d, aux: %d)\n%s\n%s"
            % (
                i + 1,
                num_words,
                tokens["source_aux"][i].count("_") + 1,
                " ".join(tokens["source"][i]),
                " ".join(tokens["source_aux"][i]),
            )
        )

        rnn_boundary = int(num_neurons / 2)
        if not is_brnn:
            rnn_boundary = num_neurons

        prev_idx = 0

        for word_idx, boundary in enumerate(word_boundaries):
            # 0 - num_neurons/2: Forward
            # num_neurons/2 - : Backward
            new_activations[word_idx, :rnn_boundary] = activations[i][
                boundary, :rnn_boundary
            ]
            if is_brnn:
                new_activations[word_idx, rnn_boundary:] = activations[i][
                    prev_idx, rnn_boundary:
                ]
            prev_idx = boundary + 1

        all_activations.append(new_activations)

    return all_activations


def sent_get_last_activations(tokens, activations):
    """Gets the summary vector for the input sentences.

    Given loaded tokens data and activations, this function picks the final token's
    activations for every sentence, essentially giving summary vectors for every
    sentence in the dataset. This is mostly applicable for RNNs.

    .. note::
        Bidirectionality is currently not handled in the case of BiRNNs.

    Parameters
    ----------
    tokens : dict
        Dictionary containing three lists, ``source``, ``source_aux`` and
        ``target``. Usually the output of ``data.loader.load_aux_data``.
    activations : list of numpy.ndarray
        Activations returned from ``loader.load_activations``.

    Returns
    -------
    activations : list of numpy.ndarray
        Summary activations corresponding to one per actual sentence in the
        original text.

    """
    all_activations = []
    num_neurons = activations[0].size(1)

    for i in tqdm(range(0, len(tokens["source"]))):
        source = tokens["source"][i]
        num_words = len(source)
        new_activations = np.zeros((1, num_neurons))

        new_activations[0, :] = activations[i][-1, :]
        all_activations.append(new_activations)

    return all_activations



""" Module for ablating neurons using various techniques.

This module provides a set of methods to ablate both layers and individual
neurons from a given set.
"""
import numpy as np

def keep_specific_neurons(X, neuron_list):
    """
    Filter activations so that they only contain specific neurons.

    .. warning::
        This function is deprecated and will be removed in future versions. Use
        ``interpretation.ablation.filter_activations_keep_neurons`` instead.

    Parameters
    ----------
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``
    neuron_list : list or numpy.ndarray
        List of neurons to keep

    Returns
    -------
    filtered_X : numpy.ndarray view
        Numpy Matrix of size [``NUM_TOKENS`` x ``len(neuron_list)``]

    """
    return X[:, neuron_list]

def filter_activations_keep_neurons(X, neurons_to_keep):
    """
    Filter activations so that they only contain specific neurons.

    .. note::
        The returned value is a view, so modifying it will modify the original
        matrix.

    Parameters
    ----------
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``
    neurons_to_keep : list or numpy.ndarray
        List of neurons to keep

    Returns
    -------
    filtered_X : numpy.ndarray view
        Numpy Matrix of size [``NUM_TOKENS`` x ``len(neurons_to_keep)``]

    """
    return X[:, neurons_to_keep]


def filter_activations_remove_neurons(X, neurons_to_remove):
    """
    Filter activations so that they do not contain specific neurons.

    .. note::
        The returned value is a view, so modifying it will modify the original
        matrix.

    Parameters
    ----------
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``
    neurons_to_remove : list or numpy.ndarray
        List of neurons to remove

    Returns
    -------
    filtered_X : numpy.ndarray view
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS - len(neurons_to_remove)``]

    """
    neurons_to_keep = np.arange(X.shape[1])
    neurons_to_keep[neurons_to_remove] = -1
    neurons_to_keep = np.where(neurons_to_keep != -1)[0]
    return X[:, neurons_to_keep]


def zero_out_activations_keep_neurons(X, neurons_to_keep):
    """
    Mask all neurons activations with zero other than specified neurons.

    Parameters
    ----------
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``
    neurons_to_keep : list or numpy.ndarray
        List of neurons to not mask

    Returns
    -------
    filtered_X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]

    """
    _X = np.zeros_like(X)

    _X[:, neurons_to_keep] = X[:, neurons_to_keep]

    return _X


def zero_out_activations_remove_neurons(X, neurons_to_remove):
    """
    Mask specific neuron activations with zero.

    Parameters
    ----------
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``
    neurons_to_remove : list or numpy.ndarray
        List of neurons to mask

    Returns
    -------
    filtered_X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]

    """
    _X = np.copy(X)

    _X[:, neurons_to_remove] = 0

    return _X

def filter_activations_by_layers(
    X, layers_to_keep, num_layers, bidirectional_filtering="none"
):
    """
    Filter activations so that they only contain specific layers.

    Useful for performing layer-wise analysis.

    Parameters
    ----------
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``
    layers_to_keep : list or numpy.ndarray
        List of layers to keep. Layers are 0-indexed
    num_layers : int
        Total number of layers in the original model.
    bidirectional_filtering : str
        Can be either "none" (Default), "forward" or "backward". Useful if the
        model being analyzed is bi-directional and only layers in a certain
        direction need to be analyzed.
    Returns
    -------
    filtered_X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS_PER_LAYER * NUM_LAYERS``]
        The second dimension is doubled if the original model is bidirectional
        and no filtering is done.

    Notes
    -----
    For bidirectional models, the method assumes that the internal structure is
    as follows: forward layer 0 neurons, backward layer 0 neurons, forward layer
    0 neurons ...

    """
    bidirectional_filtering = bidirectional_filtering.lower()
    assert bidirectional_filtering in ["none", "forward", "backward"]

    neurons_to_keep = []
    for layer in layers_to_keep:
        if bidirectional_filtering == "none":
            num_neurons_per_layer = X.shape[1] // num_layers
            start = layer * num_neurons_per_layer
            end = start + num_neurons_per_layer
        elif bidirectional_filtering == "forward":
            num_neurons_per_layer = X.shape[1] // (num_layers * 2)
            start = layer * (num_neurons_per_layer * 2)
            end = start + num_neurons_per_layer
        elif bidirectional_filtering == "backward":
            num_neurons_per_layer = X.shape[1] // (num_layers * 2)
            start = layer * num_neurons_per_layer * 2 + num_neurons_per_layer
            end = start + num_neurons_per_layer

        neurons_to_keep.append(list(range(start, end)))
    neurons_to_keep = np.concatenate(neurons_to_keep)

    return filter_activations_keep_neurons(X, neurons_to_keep)


"""Loading functions for activations, input tokens/sentences and labels

This module contains functions to load activations as well as source files with
tokens and labels. Functions that support tokenized data are also provided.
"""
import pickle
import json

import h5py
import numpy as np
import torch


def load_activations(activations_path, num_neurons_per_layer=None, is_brnn=False):
    """Load extracted activations.

    Parameters
    ----------
    activations_path : str
        Path to the activations file. Can be of type t7, pt, acts, json or hdf5
    num_neurons_per_layer : int, optional
        Number of neurons per layer - used to compute total number of layers.
        This is only necessary in the case of t7/p5/acts activations.
    is_brnn : bool, optional
        If the model used to extract activations was bidirectional (default: False)

    Returns
    -------
    activations : list of numpy.ndarray
        List of *sentence representations*, where each *sentence representation*
        is a numpy matrix of shape ``[num tokens in sentence x concatenated representation size]``
    num_layers : int
        Number of layers. This is usually representation_size/num_neurons_per_layer.
        Divide again by 2 if model was bidirectional

    """
    file_ext = activations_path.split(".")[-1]

    activations = None
    num_layers = None

    # Load activations based on type
    # Also ensure everything is on the CPU
    #   as activations may have been saved as CUDA variables
    if file_ext == "t7":
        # t7 loading requires torch < 1.0
        print("Loading seq2seq-attn activations from %s..." % (activations_path))
        assert (
            num_neurons_per_layer is not None
        ), "t7 activations require num_neurons_per_layer"
        from torch.utils.serialization import load_lua

        activations = load_lua(activations_path)["encodings"]
        activations = [a.cpu() for a in activations]
        num_layers = len(activations[0][0]) / num_neurons_per_layer
        if is_brnn:
            num_layers /= 2
    elif file_ext == "pt":
        print("Loading OpenNMT-py activations from %s..." % (activations_path))
        assert (
            num_neurons_per_layer is not None
        ), "pt activations require num_neurons_per_layer"
        activations = torch.load(activations_path)
        activations = [
            torch.stack([torch.cat(token) for token in sentence]).cpu()
            for sentence in activations
        ]
        num_layers = len(activations[0][0]) / num_neurons_per_layer
    elif file_ext == "acts":
        print("Loading generic activations from %s..." % (activations_path))
        assert (
            num_neurons_per_layer is not None
        ), "acts activations require num_neurons_per_layer"
        with open(activations_path, "rb") as activations_file:
            activations = pickle.load(activations_file)

        # Combine all layers sequentially
        print("Combining layers " + str([a[0] for a in activations]))
        activations = [a[1] for a in activations]
        num_layers = len(activations)
        num_sentences = len(activations[0])
        concatenated_activations = []
        for sentence_idx in range(num_sentences):
            sentence_acts = []
            for layer_idx in range(num_layers):
                sentence_acts.append(np.vstack(activations[layer_idx][sentence_idx]))
            concatenated_activations.append(np.concatenate(sentence_acts, axis=1))
        activations = concatenated_activations
    elif file_ext == "hdf5":
        print("Loading hdf5 activations from %s..." % (activations_path))
        representations = h5py.File(activations_path, "r")
        sentence_to_index = json.loads(representations.get("sentence_to_index")[0])
        activations = []
        # TODO: Check order
        for _, value in sentence_to_index.items():
            sentence_acts = torch.FloatTensor(representations[value])
            num_layers, sentence_length, embedding_size = (
                sentence_acts.shape[0],
                sentence_acts.shape[1],
                sentence_acts.shape[2],
            )
            num_neurons_per_layer = embedding_size
            sentence_acts = np.swapaxes(sentence_acts, 0, 1)
            sentence_acts = sentence_acts.reshape(
                sentence_length, num_layers * embedding_size
            )
            activations.append(sentence_acts.numpy())
        num_layers = len(activations[0][0]) / num_neurons_per_layer
    elif file_ext == "json":
        print("Loading json activations from %s..." % (activations_path))
        activations = []
        with open(activations_path) as fp:
            for line in fp:
                token_acts = []
                sentence_activations = json.loads(line)["features"]
                for act in sentence_activations:
                    num_neurons_per_layer = len(act["layers"][0]["values"])
                    token_acts.append(
                        np.concatenate([l["values"] for l in act["layers"]])
                    )
                activations.append(np.vstack(token_acts))

        num_layers = activations[0].shape[1] / num_neurons_per_layer
        print(len(activations), num_layers)
    else:
        assert False, "Activations must be of type t7, pt, acts, json or hdf5"

    return activations, int(num_layers)

def filter_activations_by_layers(
    train_activations, test_activations, filter_layers, rnn_size, num_layers, is_brnn
):
    """
    Filter activations so that they only contain specific layers.

    Useful for performing layer-wise analysis.

    .. warning::
        This function is deprecated and will be removed in future versions.

    Parameters
    ----------
    train_activations : list of numpy.ndarray
        List of *sentence representations* from the train set, where each
        *sentence representation* is a numpy matrix of shape
        ``[NUM_TOKENS x NUM_NEURONS]``. The method assumes that neurons from
        all layers are present, with the number of neurons in every layer given
        by ``rnn_size``
    test_activations : list of numpy.ndarray
        Similar to ``train_activations`` but with sentences from a test set.
    filter_layers : str
        A comma-separated string of the form "f1,f2,f10". "f" indicates a "forward"
        layer while "b" indicates a backword layer in a Bidirectional RNN. If the
        activations are from different kind of model, set ``is_brnn`` to ``False``
        and provide only "f" entries. The number next to "f" is the layer number,
        1-indexed. So "f1" corresponds to the embedding layer and so on.
    rnn_size : int
        Number of neurons in every layer.
    num_layers : int
        Total number of layers in the original model.
    is_brnn : bool
        Boolean indicating if the neuron activations are from a bidirectional model.

    Returns
    -------
    filtered_train_activations : list of numpy.ndarray
        Filtered train activations
    filtered_test_activations : list of numpy.ndarray
        Filtered test activations

    Notes
    -----
    For bidirectional models, the method assumes that the internal structure is
    as follows: forward layer 1 neurons, backward layer 1 neurons, forward layer
    2 neurons ...

    """
    _layers = filter_layers.split(",")

    layer_prefixes = ["f"]
    if is_brnn:
        layer_prefixes = ["f", "b"]

    # FILTER settings
    layers = list(
        range(1, num_layers + 1)
    )  # choose which layers you need the activations
    filtered_train_activations = None
    filtered_test_activations = None

    layers_idx = []
    for brnn_idx, b in enumerate(layer_prefixes):
        for l in layers:
            if "%s%d" % (b, l) in _layers:
                start_idx = brnn_idx * (num_layers * rnn_size) + (l - 1) * rnn_size
                end_idx = brnn_idx * (num_layers * rnn_size) + (l) * rnn_size

                print(
                    "Including neurons from %s%d(#%d to #%d)"
                    % (b, l, start_idx, end_idx)
                )
                layers_idx.append(np.arange(start_idx, end_idx))
    layers_idx = np.concatenate(layers_idx)

    filtered_train_activations = [a[:, layers_idx] for a in train_activations]
    filtered_test_activations = [a[:, layers_idx] for a in test_activations]

    return filtered_train_activations, filtered_test_activations


def load_aux_data(
    source_path,
    labels_path,
    source_aux_path,
    activations,
    max_sent_l,
    ignore_start_token=False,
):
    """Load word-annotated text-label pairs data represented as sentences, where
    activation extraction was performed on tokenized text. This function loads
    the source text, source tokenized text, target labels, and activations and
    tries to make them perfectly parallel, i.e. number of tokens in line N of
    source would match the number of tokens in line N of target, and number of
    tokens in source_aux will match the number of activations at index N.
    The method will delete non-matching activation/source/source_aix/target
    pairs, up to a maximum of 100 before failing. The method will also ignore
    sentences longer than the provided maximum. The activations will be modified
    in place.

    .. warning::
        This function is deprecated and will be removed in future versions.

    Parameters
    ----------
    source_path : str
        Path to the source text file, one sentence per line
    labels_path : str
        Path to the annotated labels file, one sentence per line corresponding to
        the sentences in the ``source_path`` file.
    source_aux_path : str
        Path to the source text file with tokenization, one sentence per line
    activations : list of numpy.ndarray
        Activations returned from ``loader.load_activations``
    max_sent_l : int
        Maximum length of sentences. Sentences containing more tokens will be
        ignored.
    ignore_start_token : bool, optional
        Ignore the first token. Useful if there is some line position markers
        in the source text.

    Returns
    -------
    tokens : dict
        Dictionary containing three lists, ``source``, ``source_aux`` and
        ``target``. ``source`` contains all of the sentences from``source_path``
        that were not ignored. ``source_aux`` contains all tokenized sentences
        from ``source_aux_path``. ``target`` contains the parallel set of
        annotated labels.
    """
    tokens = {"source_aux": [], "source": [], "target": []}

    skipped_lines = set()
    with open(source_aux_path) as source_aux_fp:
        for line_idx, line in enumerate(source_aux_fp):
            line_tokens = line.strip().split()
            if len(line_tokens) > max_sent_l:
                print("Skipping line #%d because of length (aux)" % (line_idx))
                skipped_lines.add(line_idx)
            if ignore_start_token:
                line_tokens = line_tokens[1:]
                activations[line_idx] = activations[line_idx][1:, :]
            tokens["source_aux"].append(line_tokens)
    with open(source_path) as source_fp:
        for line_idx, line in enumerate(source_fp):
            line_tokens = line.strip().split()
            if len(line_tokens) > max_sent_l:
                print("Skipping line #%d because of length (source)" % (line_idx))
                skipped_lines.add(line_idx)
            if ignore_start_token:
                line_tokens = line_tokens[1:]
            tokens["source"].append(line_tokens)

    with open(labels_path) as labels_fp:
        for line_idx, line in enumerate(labels_fp):
            line_tokens = line.strip().split()
            if len(line_tokens) > max_sent_l:
                print("Skipping line #%d because of length (label)" % (line_idx))
                skipped_lines.add(line_idx)
            if ignore_start_token:
                line_tokens = line_tokens[1:]
            tokens["target"].append(line_tokens)

    assert len(tokens["source_aux"]) == len(tokens["source"]) and len(
        tokens["source_aux"]
    ) == len(tokens["target"]), (
        "Number of lines do not match (source: %d, aux: %d, target: %d)!"
        % (len(tokens["source"]), len(tokens["source_aux"]), len(tokens["target"]))
    )

    assert len(activations) == len(tokens["source"]), (
        "Number of lines do not match (activations: %d, source: %d)!"
        % (len(activations), len(tokens["source"]))
    )

    for num_deleted, line_idx in enumerate(sorted(skipped_lines)):
        print("Deleting skipped line %d" % (line_idx))
        del tokens["source_aux"][line_idx]
        del tokens["source"][line_idx]
        del tokens["target"][line_idx]

    # Check if all data is well formed (whether we have activations + labels for each
    # and every word)
    invalid_activation_idx = []
    for idx, activation in enumerate(activations):
        if activation.shape[0] == len(tokens["source_aux"][idx]) and len(
            tokens["source"][idx]
        ) == len(tokens["target"][idx]):
            pass
        else:
            invalid_activation_idx.append(idx)
            print(
                "Skipping line: ",
                idx,
                "A: %d, aux: %d, src: %d, tgt: %s"
                % (
                    activation.shape[0],
                    len(tokens["source_aux"][idx]),
                    len(tokens["source"][idx]),
                    len(tokens["target"][idx]),
                ),
            )

    assert len(invalid_activation_idx) < 100, (
        "Too many mismatches (%d) - your paths are probably incorrect or something is wrong in the data!"
        % (len(invalid_activation_idx))
    )

    for num_deleted, idx in enumerate(invalid_activation_idx):
        print(
            "Deleting line %d: %d activations, %s aux, %d source, %d target"
            % (
                idx - num_deleted,
                activations[idx - num_deleted].shape[0],
                len(tokens["source_aux"][idx - num_deleted]),
                len(tokens["source"][idx - num_deleted]),
                len(tokens["target"][idx - num_deleted]),
            )
        )
        del activations[idx - num_deleted]
        del tokens["source_aux"][idx - num_deleted]
        del tokens["source"][idx - num_deleted]
        del tokens["target"][idx - num_deleted]

    for idx, activation in enumerate(activations):
        assert activation.shape[0] == len(tokens["source_aux"][idx])
        assert len(tokens["source"][idx]) == len(tokens["target"][idx])

    return tokens


def load_data(
    source_path,
    labels_path,
    activations,
    max_sent_l,
    ignore_start_token=False,
    sentence_classification=False,
):
    """Load word-annotated text-label pairs data represented as sentences. This
    function loads the source text, target labels, and activations and tries to
    make them perfectly parallel, i.e. number of tokens in line N of source would
    match the number of tokens in line N of target, and also match the number of
    activations at index N. The method will delete non-matching activation/source/target
    pairs, up to a maximum of 100 before failing. The method will also ignore
    sentences longer than the provided maximum. The activations will be modified
    in place.

    Parameters
    ----------
    source_path : str
        Path to the source text file, one sentence per line
    labels_path : str
        Path to the annotated labels file, one sentence per line corresponding to
        the sentences in the ``source_path`` file.
    activations : list of numpy.ndarray
        Activations returned from ``loader.load_activations``
    max_sent_l : int
        Maximum length of sentences. Sentences containing more tokens will be
        ignored.
    ignore_start_token : bool, optional
        Ignore the first token. Useful if there is some line position markers
        in the source text.
    sentence_classification : bool, optional
        Flag to indicate if this is a sentence classification task, where every
        sentence actually has only a single activation (e.g. [CLS] token's
        activations in the case of BERT)

    Returns
    -------
    tokens : dict
        Dictionary containing two lists, ``source`` and ``target``. ``source``
        contains all of the sentences from ``source_path`` that were not ignored.
        ``target`` contains the parallel set of annotated labels.

    """
    tokens = {"source": [], "target": []}

    with open(source_path) as source_fp:
        for line_idx, line in enumerate(source_fp):
            line_tokens = line.strip().split()
            if len(line_tokens) > max_sent_l:
                continue
            if ignore_start_token:
                line_tokens = line_tokens[1:]
                activations[line_idx] = activations[line_idx][1:, :]
            tokens["source"].append(line_tokens)

    with open(labels_path) as labels_fp:
        for line in labels_fp:
            line_tokens = line.strip().split()
            if len(line_tokens) > max_sent_l:
                continue
            if ignore_start_token:
                line_tokens = line_tokens[1:]
            tokens["target"].append(line_tokens)

    assert len(tokens["source"]) == len(tokens["target"]), (
        "Number of lines do not match (source: %d, target: %d)!"
        % (len(tokens["source"]), len(tokens["target"]))
    )

    assert len(activations) == len(tokens["source"]), (
        "Number of lines do not match (activations: %d, source: %d)!"
        % (len(activations), len(tokens["source"]))
    )

    # Check if all data is well formed (whether we have activations + labels for
    # each and every word)
    invalid_activation_idx = []
    for idx, activation in enumerate(activations):
        if activation.shape[0] == len(tokens["source"][idx]) and (
            sentence_classification or activation.shape[0] == len(tokens["target"][idx])
        ):
            pass
        else:
            invalid_activation_idx.append(idx)
            print("Skipping line: ", idx)
            print(
                "A: %d, S: %d, T: %d"
                % (
                    activation.shape[0],
                    len(tokens["source"][idx]),
                    len(tokens["target"][idx]),
                )
            )

    assert len(invalid_activation_idx) < 100, (
        "Too many mismatches (%d) - your paths are probably incorrect or something is wrong in the data!"
        % (len(invalid_activation_idx))
    )

    for num_deleted, idx in enumerate(invalid_activation_idx):
        print(
            "Deleting line %d: %d activations, %d source, %d target"
            % (
                idx - num_deleted,
                activations[idx - num_deleted].shape[0],
                len(tokens["source"][idx - num_deleted]),
                len(tokens["target"][idx - num_deleted]),
            )
        )
        del activations[idx - num_deleted]
        del tokens["source"][idx - num_deleted]
        del tokens["target"][idx - num_deleted]

    for idx, activation in enumerate(activations):
        assert activation.shape[0] == len(tokens["source"][idx])
        if not sentence_classification:
            assert activation.shape[0] == len(tokens["target"][idx])

    # TODO: Return activations
    return tokens


def load_sentence_data(source_path, labels_path, activations):
    """Loads sentence-annotated text-label pairs. This function loads the source
    text, target labels, and activations and tries to make them perfectly
    parallel, i.e. number of tokens in line N of source would
    match the number of activations at index N. The method will delete
    non-matching activation/source pairs. The activations will be modified
    in place.

    Parameters
    ----------
    source_path : str
        Path to the source text file, one sentence per line
    labels_path : str
        Path to the annotated labels file, one sentence per line corresponding to
        the sentences in the ``source_path`` file.
    activations : list of numpy.ndarray
        Activations returned from ``loader.load_activations``

    Returns
    -------
    tokens : dict
        Dictionary containing two lists, ``source`` and ``target``. ``source``
        contains all of the sentences from ``source_path`` that were not ignored.
        ``target`` contains the parallel set of annotated labels.

    """
    tokens = {"source": [], "target": []}

    with open(source_path) as source_fp:
        for line_idx, line in enumerate(source_fp):
            tokens["source"].append(["sentence_%d" % (line_idx)])

    with open(labels_path) as labels_fp:
        for line in labels_fp:
            line_tokens = line.strip().split()
            tokens["target"].append(line_tokens)

    assert len(tokens["source"]) == len(tokens["target"]), (
        "Number of lines do not match (source: %d, target: %d)!"
        % (len(tokens["source"]), len(tokens["target"]))
    )

    assert len(activations) == len(tokens["source"]), (
        "Number of lines do not match (activations: %d, source: %d)!"
        % (len(activations), len(tokens["source"]))
    )

    # Check if all data is well formed (whether we have activations + labels for
    # each and every word)
    for idx, activation in enumerate(activations):
        assert activation.shape[0] == len(tokens["source"][idx])

    return tokens



"""Module for corpus based analysis.

This module contains functions that relate neurons to corpus elements like
words and sentences

"""
from collections import defaultdict

import numpy as np


def get_top_words(tokens, activations, neuron, num_tokens=0):
    """
    Get top activating words for any given neuron.

    This method compares the activations of the given neuron across all tokens,
    and extracts tokens that account for the largest variance for that given
    neuron. It also returns a normalized score for each token, depicting their
    contribution to the overall variance.

    Parameters
    ----------
    tokens : dict
        Dictionary containing atleast one list with the key ``source``. Usually
        returned from ``data.loader.load_data``
    activations : list of numpy.ndarray
        List of *sentence representations*, where each *sentence representation*
        is a numpy matrix of shape
        ``[num tokens in sentence x concatenated representation size]``. Usually
        retured from ``data.loader.load_activations``
    neuron : int
        Index of the neuron relative to ``X``
    num_tokens: int, optional
        Number of top tokens to return. Defaults to 0, which returns all tokens
        with a non-neglible contribution to the variance

    Returns
    -------
    top_neurons : list of tuples
        List of tuples, where each tuple is a (token, score) element

    """
    MIN_THRESHOLD = 0.1  # threshold for "negligible activation score"

    activation_values = [
        sentence_activations[:, neuron] for sentence_activations in activations
    ]
    activation_values = np.concatenate(activation_values)

    tokens = [token for sentence in tokens["source"] for token in sentence]
    mean = np.mean(activation_values)
    std = np.std(activation_values)

    token_wise_scores = np.abs((activation_values - mean) / std)
    type_wise_scores_aggregation = defaultdict(lambda: (0, 0))
    for idx, token in enumerate(tokens):
        curr_sum, curr_count = type_wise_scores_aggregation[token]
        type_wise_scores_aggregation[token] = (
            curr_sum + token_wise_scores[idx],
            curr_count + 1,
        )

    # Normalize by count
    type_wise_scores = [
        (k, v[0] / v[1]) for k, v in type_wise_scores_aggregation.items()
    ]

    # Normalize scores by max
    max_score = max([s for _, s in type_wise_scores])
    type_wise_scores = [(k, v / max_score) for k, v in type_wise_scores]

    # Sort and filter scores
    sorted_types_scores = sorted(type_wise_scores, key=lambda x: -x[1])
    sorted_types_scores = [
        (k, v) for (k, v) in sorted_types_scores if v > MIN_THRESHOLD
    ]

    if num_tokens > 0:
        sorted_types_scores = sorted_types_scores[:num_tokens]

    return sorted_types_scores



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

sns.set(style="darkgrid")
sns.set_palette("tab10")

### Plotting
def plot_accuracies_per_tag(title, **kwargs):
    if kwargs is None:
        print("Need accuracies to plot")
        return

    classes = set(sum([list(kwargs[exp].keys()) for exp in kwargs], []))
    classes = list(classes)
    classes.remove('__OVERALL__')
    classes = ['Overall'] + sorted(classes)

    experiments = {}
    for exp in kwargs:
        experiments[exp] = {k: kwargs[exp][k] for k in kwargs[exp] if k != '__OVERALL__'}
        experiments[exp]['Overall'] = kwargs[exp]['__OVERALL__']

    df = pd.DataFrame([{
            'experiment': exp,
            **experiments[exp]
        } for exp in experiments])

    df = pd.melt(df, id_vars="experiment", var_name="type", value_name="accuracy")

    fig, ax = plt.subplots(1,1, figsize=(16,6))
    plt.sca(ax)
    ax.tick_params(axis='x', rotation=90)
    ax.set_ylim((0.5, 1.05))
    sns.factorplot(ax=ax, x='type', y='accuracy', hue='experiment', 
                    data=df, kind='bar', order=classes, legend_out=True)
    
    handles, labels = ax.get_legend_handles_labels()
    l = ax.legend(handles=handles, labels=labels, title=title, frameon=True, fancybox=True, framealpha=0.8, facecolor="#FFFFFF")

    return fig

def plot_distributedness(title, top_neurons_per_tag):
    fig, ax = plt.subplots(1,1, figsize=(16,6))

    tags = sorted(top_neurons_per_tag.keys())
    number_of_neurons = [len(top_neurons_per_tag[tag]) for tag in tags]
    ax = sns.barplot(x=tags, y=number_of_neurons, color="yellowgreen")
    ax.tick_params(axis='x', rotation=90)
    ax.set_ylabel("Number of Important Neurons")
    ax.set_title(title)

def plot_accuracies(title, overall_acc,
        top_10_acc, random_10_acc, bottom_10_acc,
        top_15_acc, random_15_acc, bottom_15_acc,
        top_20_acc, random_20_acc, bottom_20_acc):
    exps = ["Full", " ",
        "Top 10%", "Random 10%", "Bottom 10%", "  ",
        "Top 15%", "Random 15%", "Bottom 15%", "   ",
        "Top 20%", "Random 20%", "Bottom 20%"]

    concise_exps = ["", "Full", " ",
        "  ", "10% Neurons", "   ", "    ",
        "     ", "15% Neurons", "      ", "       ",
        "        ", "20% Neurons", "         ", "          "]

    accs = np.array([0, overall_acc, 0,
        top_10_acc, random_10_acc, bottom_10_acc, 0, 
        top_15_acc, random_15_acc, bottom_15_acc, 0,
        top_20_acc, random_20_acc, bottom_20_acc, 0])

    min_value = np.min(accs[accs>0.01])

    colors = [(0,0,0), 'limegreen', (0,0,0),
            'dodgerblue', 'gold', 'orangered', 'g',
            'dodgerblue', 'gold', 'orangered', 'g',
            'dodgerblue', 'gold', 'orangered', 'g', (0,0,0)]
    ax = sns.barplot(x=concise_exps, y=accs, palette=colors)

    ax.set_ylim((max(min_value-0.1,0), 1.04))

    ax.set_title(title)
    handles = [ax.patches[1], ax.patches[3], ax.patches[4], ax.patches[5]]
    labels = ["All Neurons", "Top Neurons", "Random Neurons", "Bottom Neurons"]
    ax.legend(handles=handles, labels=labels, frameon=True, fancybox=True, ncol=2, framealpha=0.8, facecolor="#FFFFFF")



import os
import sys

import svgwrite

from neurox.data.extraction.transformers_extractor import (
    get_model_and_tokenizer,
    extract_sentence_representations,
)


# Helper methods
def _break_lines(text, limit=50):
    lines = []
    curr_line = ""
    for token in text.split(" "):
        if len(curr_line) + 1 + len(token) <= limit:
            curr_line += token + " "
        else:
            lines.append(curr_line[:-1])
            curr_line = token + " "
    lines.append(curr_line[:-1])
    return lines


def _get_rect_style(color, opacity):
    return """opacity:%0.5f;
            fill:%s;
            fill-opacity:1;
            stroke:none;
            stroke-width:0.26499999;
            stroke-linecap:round;
            stroke-linejoin:miter;
            stroke-miterlimit:4;
            stroke-dasharray:none;
            stroke-dashoffset:0;
            stroke-opacity:1""" % (
        opacity,
        color,
    )


def _get_text_style(font_size):
    return f"""font-style: normal;
            font-variant: normal;
            font-weight: normal;
            font-stretch: normal;
            font-size: {font_size:0.2f}px;
            line-height: 125%;
            font-family: "Courier";
            -inkscape-font-specification: "Courier";
            font-variant-ligatures: none;
            font-variant-caps: normal;
            font-variant-numeric: normal;
            text-align: start;
            writing-mode: lr-tb;
            text-anchor: start;
            stroke-width: 0.26458332px"""


MARGIN = 10


def visualize_activations(
    tokens,
    activations,
    darken=2,
    colors=["#d35f5f", "#00aad4"],
    text_direction="ltr",
    char_limit=60,
    font_size=20,
    filter_fn=lambda x: x
):
    """
    Visualize activation values for a particular neuron on some text.

    This method returns an SVG drawing of text with every token's background
    color set according to the passed in activation values (red for negative
    values and blue for positive).

    Parameters
    ----------
    tokens : list of str
        List of tokens over which the activations have been computed. In the
        rendered image, tokens will be separated by a single space.
    activations: list of float
        List of activation values, one per token.
    darken : int, optional
        Number of times to render the red/blue background. Increasing this
        value will reduce contrast but may help in better distinguishing between
        tokens. Defaults to 2
    colors : list of str, optional
        List of two elements, the first indicating the color of the lowest
        activation value and the second indicating the color of the highest
        activation value. Defaults to shades of red and blue respectively
    text_direction : str, optional
        One of ``ltr`` or ``rtl``, indicating if the language being rendered is
        written left to right or right to left. Defaults to ``ltr``
    char_limit : int, optional
        Maximum number of characters per line. Defaults to 60
    font_size : int, optional
        Font size in pixels. Defaults to 20px
    filter_fn : str or fn, optional
        Additional functiont that modifies the incoming activations. Defaults to
        None resulting in keeping the activations as is. If fn is provided, it
        must accept a list of activations and return a list of exactly the same
        number of elements. str choices are currently:
            'top_tokens': Only highlights tokens whose activation values are within
                80% of the top activating token in a given sentence. Absolute values
                are used for comparison.

    Returns
    -------
    rendered_svg : svgwrite.Drawing
        A SVG object that you can either save to file, convert into a png within
        python using an external library like Pycairo, or display in a notebook
        using the ``display`` from the module ``IPython.display``
    """
    ################################ Validation ################################
    valid_text_directions = ["ltr", "rtl"]
    text_direction = text_direction.lower()
    assert (
        text_direction in valid_text_directions
    ), f"text_direction must be one of {valid_text_directions}"

    assert len(tokens) == len(
        activations
    ), f"Number of tokens and activations must match"

    ################################ Filtering ################################
    if filter_fn == "top_tokens":
        def keep_top_tokens(acts):
            max_val = max([abs(a) for a in acts])
            new_acts = [a if abs(a) > 0.8 * max_val else 0 for a in acts]
            return new_acts
        filter_fn = keep_top_tokens
    activations_filtered = filter_fn(activations)
    assert len(activations) == len(activations_filtered)
    activations = activations_filtered

    ############################## Drawing Setup ###############################
    text = " ".join(tokens)

    # Estimate individual character sizes
    char_width = font_size * 0.601  # Magic number for Courier font
    char_height = font_size * 1.25  # 1.25 is line height of rendered font

    # Compute number of lines
    lines = _break_lines(text, limit=char_limit)

    # Compute image size based on character sizes and number of lines
    image_height = len(lines) * char_height * 1.2
    image_width = (char_limit + 1) * char_width

    # Create drawing canvas
    dwg = svgwrite.Drawing("tmp.svg", size=(image_width, image_height), profile="full")
    dwg.viewbox(0, 0, image_width, image_height)
    group = dwg.g()

    ####################### Activation Rendering limits ########################
    scores = activations
    max_score = max(scores)
    min_score = abs(min(scores))
    limit = max(max_score, min_score)

    for _ in range(darken):
        word_idx = 0
        line_horizontal_offsets = []
        for line_idx, line in enumerate(lines):
            char_idx = 0
            words = line.split(" ")
            if text_direction == "rtl":
                words = reversed(words)
            for word in words:
                score = scores[word_idx]
                if score > 0:
                    color = colors[1]
                    opacity = score / limit
                else:
                    color = colors[0]
                    opacity = abs(score) / limit

                # Add rectangle for every character in current word
                for _ in word:
                    rect_position = (char_idx * char_width, 7 + line_idx * char_height)
                    rect_size = (f"{char_width:0.3f}px", f"{char_height:0.3f}px")
                    group.add(
                        dwg.rect(
                            insert=rect_position,
                            size=rect_size,
                            style=_get_rect_style(color, opacity),
                        )
                    )
                    char_idx += 1

                # Add rectangle for empty space after word
                final_rect_pos = (char_idx * char_width, 7 + line_idx * char_height)
                final_rect_size = (f"{char_width:0.3f}px", f"{char_height:0.3f}px")
                group.add(
                    dwg.rect(
                        insert=final_rect_pos,
                        size=final_rect_size,
                        style=_get_rect_style(color, opacity),
                    )
                )

                char_idx += 1
                word_idx += 1
            if text_direction == "ltr":
                line_horizontal_offsets.append(0)
            else:
                line_horizontal_offsets.append(char_idx * char_width)

        # Draw the actual text over the drawn rectangles
        for line_idx, line in enumerate(lines):
            text_insert = (
                line_horizontal_offsets[line_idx],
                font_size * 1.25 * (line_idx + 1),
            )
            text = dwg.text(
                line, insert=text_insert, fill="black", style=_get_text_style(font_size)
            )
            group.add(text)

    dwg.add(group)

    return dwg


class TransformersVisualizer:
    """
    Helper class to visualize sentences using activations from a
    ``transformers`` model.

    Attributes
    ----------
    model_name : str
        A ``transformers`` model name or path, e.g. ``bert-base-uncased``
    model : ``transformers`` model
        The loaded model
    tokenizer : ``transformers`` tokenizer
        The loaded tokenizer

    Methods
    -------
    __call__(tokens, layer, neuron)
        An object of this class can be called directly to get the visualized
        activations

    Examples
    --------
    >>> visualizer = TransformersVisualizer('bert-base-uncased')
    >>> svg1 = visualizer(["This", "is", "a", "test"], 0, 10)
    >>> svg2 = visualizer(["This", "is", "another", "test"], 5, 767)

    """

    def __init__(self, model_name):
        "Load the model and tokenizer"
        self.model_name = model_name
        self.model, self.tokenizer = get_model_and_tokenizer(model_name)

    class __HiddenPrints__:
        def __enter__(self):
            self._original_stdout = sys.stdout
            sys.stdout = open(os.devnull, "w")

        def __exit__(self, exc_type, exc_val, exc_tb):
            sys.stdout.close()
            sys.stdout = self._original_stdout

    def __call__(self, tokens, layer, neuron, **kwargs):
        """
        Visualize the activations of ``neuron`` from ``layer`` in the loaded
        model on the given tokens.

        Parameters
        ----------
        tokens : list of str
            List of tokens to compute and visualize the activations for
        layer : int
            Layer index of the chosen neuron to visualize in the loaded model
        neuron : int
            Neuron index of the chosen neuron
        kwargs : dict
            Dictionary of keyword arguments to pass to ``visualize_activations``
        """

        with self.__HiddenPrints__():
            activations, _ = extract_sentence_representations(
                " ".join(tokens), self.model, self.tokenizer
            )
        return visualize_activations(tokens, activations[layer, :, neuron], **kwargs)



"""
Module that wraps around several standard metrics
"""
import numpy as np

from scipy.stats import pearsonr, spearmanr
from sklearn.metrics import matthews_corrcoef as mcc, f1_score


def _numpyfy(x):
    """
    Internal helper function to convert standard lists into numpy arrays.

    Parameters
    ----------
    x : list or numpy.ndarray
        A list of numbers

    Returns
    -------
    x : numpy.ndarray
        Numpy array with the original numbers

    """
    if isinstance(x, np.ndarray):
        return x
    return np.array(x)


def accuracy(preds, labels):
    """
    Accuracy.

    Parameters
    ----------
    preds : list or numpy.ndarray
        A list of predictions from a model
    labels : list or numpy.ndarray
        A list of ground truth labels with the same number of elements as
        ``preds``

    Returns
    -------
    accuracy : float
        Accuracy of the model
    """
    preds = _numpyfy(preds)
    labels = _numpyfy(labels)
    return (preds == labels).mean()

import json
def f1(preds, labels):
    """
    F-Score or F1 score.

    .. note::
        The implementation from ``sklearn.metrics`` is used to compute the score.

    Parameters
    ----------
    preds : list or numpy.ndarray
        A list of predictions from a model
    labels : list or numpy.ndarray
        A list of ground truth labels with the same number of elements as
        ``preds``

    Returns
    -------
    f1_score : float
        F-Score of the model
    """
    preds = _numpyfy(preds)
    labels = _numpyfy(labels)
    return f1_score(y_true=labels, y_pred=preds)


def accuracy_and_f1(preds, labels):
    """
    Mean of Accuracy and F-Score.

    .. note::
        The implementation from ``sklearn.metrics`` is used to compute the
        F-Score.

    Parameters
    ----------
    preds : list or numpy.ndarray
        A list of predictions from a model
    labels : list or numpy.ndarray
        A list of ground truth labels with the same number of elements as
        ``preds``

    Returns
    -------
    acc_f1_mean : float
        Mean of Accuracy and F-Score of the model
    """
    preds = _numpyfy(preds)
    labels = _numpyfy(labels)
    acc = accuracy(preds, labels)
    f1_s = f1_score(y_true=labels, y_pred=preds)
    return (acc + f1_s) / 2


def pearson(preds, labels):
    """
    Pearson's correlation coefficient

    .. note::
        The implementation from ``scipy.stats`` is used to compute the score.

    Parameters
    ----------
    preds : list or numpy.ndarray
        A list of predictions from a model
    labels : list or numpy.ndarray
        A list of ground truth labels with the same number of elements as
        ``preds``

    Returns
    -------
    pearson_score : float
        Pearson's correlation coefficient of the model
    """
    preds = _numpyfy(preds)
    labels = _numpyfy(labels)
    return pearsonr(preds, labels)[0]


def spearman(preds, labels):
    """
    Spearman correlation coefficient

    .. note::
        The implementation from ``scipy.stats`` is used to compute the score.

    Parameters
    ----------
    preds : list or numpy.ndarray
        A list of predictions from a model
    labels : list or numpy.ndarray
        A list of ground truth labels with the same number of elements as
        ``preds``

    Returns
    -------
    spearman_score : float
        Spearman correlation coefficient of the model
    """
    preds = _numpyfy(preds)
    labels = _numpyfy(labels)
    return spearmanr(preds, labels)[0]


def pearson_and_spearman(preds, labels):
    """
    Mean of Pearson and Spearman correlation coefficients.

    .. note::
        The implementation from ``scipy.stats`` is used to compute the scores.

    Parameters
    ----------
    preds : list or numpy.ndarray
        A list of predictions from a model
    labels : list or numpy.ndarray
        A list of ground truth labels with the same number of elements as
        ``preds``

    Returns
    -------
    pearson_spearman_mean : float
        Mean of Pearson and Spearman correlation coefficients of the model
    """
    preds = _numpyfy(preds)
    labels = _numpyfy(labels)
    return (pearson(preds, labels) + spearman(preds, labels)) / 2


def matthews_corrcoef(preds, labels):
    """
    Matthew's correlation coefficient

    .. note::
        The implementation from ``sklearn.metrics`` is used to compute the score.

    Parameters
    ----------
    preds : list or numpy.ndarray
        A list of predictions from a model
    labels : list or numpy.ndarray
        A list of ground truth labels with the same number of elements as
        ``preds``

    Returns
    -------
    mcc_score : float
        Matthew's correlation coefficient of the model
    """
    preds = _numpyfy(preds)
    labels = _numpyfy(labels)
    return mcc(preds, labels)


def compute_score(preds, labels, metric):
    """
    Utility function to compute scores using several metrics.

    Parameters
    ----------
    preds : list or numpy.ndarray
        A list of predictions from a model
    labels : list or numpy.ndarray
        A list of ground truth labels with the same number of elements as
        ``preds``
    metric : str
        One of ``accuracy``, ``f1``, ``accuracy_and_f1``, ``pearson``,
        ``spearman``, ``pearson_and_spearman`` or ``matthews_corrcoef``.

    Returns
    -------
    score : float
        Score of the model with the chosen metric
    """
    if metric == 'accuracy':
        return accuracy(preds, labels)
    elif metric == 'f1':
        return f1(preds, labels)
    elif metric == 'accuracy_and_f1':
        return accuracy_and_f1(preds, labels)
    elif metric == 'pearson':
        return pearson(preds, labels)
    elif metric == 'spearman':
        return spearman(preds, labels)
    elif metric == 'pearson_and_spearman':
        return pearson_and_spearman(preds, labels)
    elif metric == 'matthews_corrcoef':
        return matthews_corrcoef(preds, labels)


"""Representations Extractor for ``transformers`` toolkit models.

Script that given a file with input sentences and a ``transformers``
model, extracts representations from all layers of the model. The script
supports aggregation over sub-words created due to the tokenization of
the provided model.

Author: Fahim Dalvi
Last Modified: 2 March, 2020
Last Modified: 9 September, 2020
Last Modified: 15 September, 2020
Last Modified: 1 February, 2020
"""

import argparse
import collections
import json
import sys

import numpy as np
import torch
import h5py

from tqdm import tqdm
from transformers import AutoTokenizer, AutoModel


def get_model_and_tokenizer(model_desc, device="cpu", random_weights=False):
    """
    Automatically get the appropriate ``transformers`` model and tokenizer based
    on the model description

    Parameters
    ----------
    model_desc : str
        Model description; can either be a model name like ``bert-base-uncased``,
        a comma separated list indicating <model>,<tokenizer> (since 1.0.8),
        or a path to a trained model

    device : str, optional
        Device to load the model on, cpu or gpu. Default is cpu.

    random_weights : bool, optional
        Whether the weights of the model should be randomized. Useful for analyses
        where one needs an untrained model.

    Returns
    -------
    model : transformers model
        An instance of one of the transformers.modeling classes
    tokenizer : transformers tokenizer
        An instance of one of the transformers.tokenization classes
    """
    model_desc = model_desc.split(",")
    if len(model_desc) == 1:
        model_name = model_desc[0]
        tokenizer_name = model_desc[0]
    else:
        model_name = model_desc[0]
        tokenizer_name = model_desc[1]
    model = AutoModel.from_pretrained(model_name, output_hidden_states=True).to(device)
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_name)

    if random_weights:
        print("Randomizing weights")
        model.init_weights()

    return model, tokenizer


def aggregate_repr(state, start, end, aggregation):
    """
    Function that aggregates activations/embeddings over a span of subword tokens.
    This function will usually be called once per word. For example, if we had the sentence::

        This is an example

    which is tokenized by BPE into::

        this is an ex @@am @@ple

    The function should be called 4 times::

        aggregate_repr(state, 0, 0, aggregation)
        aggregate_repr(state, 1, 1, aggregation)
        aggregate_repr(state, 2, 2, aggregation)
        aggregate_repr(state, 3, 5, aggregation)

    Returns a zero vector if end is less than start, i.e. the request is to
    aggregate over an empty slice.

    Parameters
    ----------
    state : numpy.ndarray
        Matrix of size [ NUM_LAYERS x NUM_SUBWORD_TOKENS_IN_SENT x LAYER_DIM]
    start : int
        Index of the first subword of the word being processed
    end : int
        Index of the last subword of the word being processed
    aggregation : {'first', 'last', 'average'}
        Aggregation method for combining subword activations

    Returns
    -------
    word_vector : numpy.ndarray
        Matrix of size [NUM_LAYERS x LAYER_DIM]
    """
    if end < start:
        sys.stderr.write("WARNING: An empty slice of tokens was encountered. " +
            "This probably implies a special unicode character or text " +
            "encoding issue in your original data that was dropped by the " +
            "transformer model's tokenizer.\n")
        return np.zeros((state.shape[0], state.shape[2]))
    if aggregation == "first":
        return state[:, start, :]
    elif aggregation == "last":
        return state[:, end, :]
    elif aggregation == "average":
        return np.average(state[:, start : end + 1, :], axis=1)


def extract_sentence_representations(
    sentence,
    model,
    tokenizer,
    device="cpu",
    include_embeddings=True,
    aggregation="last",
    tokenization_counts={},
    include_special_tokens=True
):
    """
    Get representations for one sentence
    """
    # this follows the HuggingFace API for transformers

    special_tokens = [
        x for x in tokenizer.all_special_tokens if x != tokenizer.unk_token
    ]
    special_tokens_ids = tokenizer.convert_tokens_to_ids(special_tokens)

    original_tokens = sentence.split(" ")
    # Add a letter and space before each word since some tokenizers are space sensitive
    tmp_tokens = [
        "a" + " " + x if x_idx != 0 else x for x_idx, x in enumerate(original_tokens)
    ]
    assert len(original_tokens) == len(tmp_tokens)

    with torch.no_grad():
        # Get tokenization counts if not already available
        for token_idx, token in enumerate(tmp_tokens):
            tok_ids = [
                x for x in tokenizer.encode(token) if x not in special_tokens_ids
            ]
            if token_idx != 0:
                # Ignore the first token (added letter)
                tok_ids = tok_ids[1:]

            if token in tokenization_counts:
                assert tokenization_counts[token] == len(
                    tok_ids
                ), "Got different tokenization for already processed word"
            else:
                tokenization_counts[token] = len(tok_ids)
        ids = tokenizer.encode(sentence, truncation=True)
        input_ids = torch.tensor([ids]).to(device)
        # Hugging Face format: tuple of torch.FloatTensor of shape (batch_size, sequence_length, hidden_size)
        # Tuple has 13 elements for base model: embedding outputs + hidden states at each layer
        all_hidden_states = model(input_ids)[-1]

        if include_embeddings:
            all_hidden_states = [
                hidden_states[0].cpu().numpy() for hidden_states in all_hidden_states
            ]
        else:
            all_hidden_states = [
                hidden_states[0].cpu().numpy()
                for hidden_states in all_hidden_states[1:]
            ]
        all_hidden_states = np.array(all_hidden_states)

    print('Sentence         : "%s"' % (sentence))
    print("Original    (%03d): %s" % (len(original_tokens), original_tokens))
    print(
        "Tokenized   (%03d): %s"
        % (
            len(tokenizer.convert_ids_to_tokens(ids)),
            tokenizer.convert_ids_to_tokens(ids),
        )
    )

    assert all_hidden_states.shape[1] == len(ids)

    # Handle special tokens
    

    filtered_ids = ids

    idx_special_tokens = [
        t_i for t_i, x in enumerate(ids) if x in special_tokens_ids
    ]
    special_token_ids = [ids[t_i] for t_i in idx_special_tokens]
    
    if not include_special_tokens:
        idx_without_special_tokens = [
            t_i for t_i, x in enumerate(ids) if x not in special_tokens_ids
        ]
        filtered_ids = [ids[t_i] for t_i in idx_without_special_tokens]
        all_hidden_states = all_hidden_states[:, idx_without_special_tokens, :]
        assert all_hidden_states.shape[1] == len(filtered_ids)
    print(
        "Filtered   (%03d): %s"
        % (
            len(tokenizer.convert_ids_to_tokens(filtered_ids)),
            tokenizer.convert_ids_to_tokens(filtered_ids),
        )
    )
    segmented_tokens = tokenizer.convert_ids_to_tokens(filtered_ids)

    # Perform actual subword aggregation/detokenization
    counter = 0
    detokenized = []
    final_hidden_states = np.zeros(
        (all_hidden_states.shape[0], len(original_tokens) + len(ids) - len(filtered_ids), all_hidden_states.shape[2])
    )
    inputs_truncated = False

    if include_special_tokens:
        last_special_token_pointer = 0
    for token_idx, token in enumerate(tmp_tokens):
        # Handle special tokens
        if include_special_tokens and last_special_token_pointer < len(idx_special_tokens):
            if counter == idx_special_tokens[last_special_token_pointer]:
                print(f"Found {segmented_tokens[idx_special_tokens[last_special_token_pointer]]}")
                last_special_token_pointer += 1
                counter += 1

        current_word_start_idx = counter
        current_word_end_idx = counter + tokenization_counts[token]

        # Check for truncated hidden states in the case where the
        # original word was actually tokenized
        if  (tokenization_counts[token] != 0 and current_word_start_idx >= all_hidden_states.shape[1]) \
                or current_word_end_idx > all_hidden_states.shape[1]:
            final_hidden_states = final_hidden_states[:, :len(detokenized), :]
            inputs_truncated = True
            break

        final_hidden_states[:, len(detokenized), :] = aggregate_repr(
            all_hidden_states,
            current_word_start_idx,
            current_word_end_idx - 1,
            aggregation,
        )
        detokenized.append(
            "".join(segmented_tokens[current_word_start_idx:current_word_end_idx])
        )
        counter += tokenization_counts[token]

    if include_special_tokens:
        while counter < len(segmented_tokens):
            if last_special_token_pointer >= len(idx_special_tokens):
                break
        
            if counter == idx_special_tokens[last_special_token_pointer]:
                print(f"Found {segmented_tokens[idx_special_tokens[last_special_token_pointer]]}")
                last_special_token_pointer += 1
                counter += 1

    print("Detokenized (%03d): %s" % (len(detokenized), detokenized))
    print("Counter: %d" % (counter))

    if inputs_truncated:
        print("WARNING: Input truncated because of length, skipping check")
    else:
        assert counter == len(filtered_ids)
        assert len(detokenized) == len(original_tokens)
    print("===================================================================")

    return final_hidden_states, detokenized


def extract_representations(
    model_desc,
    input_corpus,
    output_file,
    device="cpu",
    aggregation="last",
    output_type="json",
    random_weights=False,
    ignore_embeddings=False,
):
    print(f"Loading model: {model_desc}")
    model, tokenizer = get_model_and_tokenizer(
        model_desc, device=device, random_weights=random_weights
    )

    print("Reading input corpus")

    def corpus_generator(input_corpus_path):
        with open(input_corpus_path, "r") as fp:
            for line in fp:
                yield line.strip()
            return

    print("Preparing output file")
    if output_type == "hdf5":
        if not output_file.endswith(".hdf5"):
            print(
                "[WARNING] Output filename (%s) does not end with .hdf5, but output file type is hdf5."
                % (output_file)
            )
        output_file = h5py.File(output_file, "w")
        sentence_to_index = {}
    elif output_type == "json":
        if not output_file.endswith(".json"):
            print(
                "[WARNING] Output filename (%s) does not end with .json, but output file type is json."
                % (output_file)
            )
        output_file = open(output_file, "w", encoding="utf-8")

    print("Extracting representations from model")
    tokenization_counts = {} # Cache for tokenizer rules
    for sentence_idx, sentence in enumerate(corpus_generator(input_corpus)):
        hidden_states, extracted_words = extract_sentence_representations(
            sentence,
            model,
            tokenizer,
            device=device,
            include_embeddings=(not ignore_embeddings),
            aggregation=aggregation,
            tokenization_counts=tokenization_counts
        )

        print("Hidden states: ", hidden_states.shape)
        print("# Extracted words: ", len(extracted_words))

        if output_type == "hdf5":
            output_file.create_dataset(
                str(sentence_idx),
                hidden_states.shape,
                dtype="float32",
                data=hidden_states,
            )
            # TODO: Replace with better implementation with list of indices
            final_sentence = sentence
            counter = 1
            while final_sentence in sentence_to_index:
                counter += 1
                final_sentence = f"{sentence} (Occurrence {counter})"
            sentence = final_sentence
            sentence_to_index[sentence] = str(sentence_idx)
        elif output_type == "json":
            output_json = collections.OrderedDict()
            output_json["linex_index"] = sentence_idx
            all_out_features = []

            for word_idx, extracted_word in enumerate(extracted_words):
                all_layers = []
                for layer_idx in range(hidden_states.shape[0]):
                    layers = collections.OrderedDict()
                    layers["index"] = layer_idx
                    layers["values"] = [
                        round(x.item(), 8)
                        for x in hidden_states[layer_idx, word_idx, :]
                    ]
                    all_layers.append(layers)
                out_features = collections.OrderedDict()
                out_features["token"] = extracted_word
                out_features["layers"] = all_layers
                all_out_features.append(out_features)
            output_json["features"] = all_out_features
            output_file.write(json.dumps(output_json) + "\n")

    if output_type == "hdf5":
        sentence_index_dataset = output_file.create_dataset(
            "sentence_to_index", (1,), dtype=h5py.special_dtype(vlen=str)
        )
        sentence_index_dataset[0] = json.dumps(sentence_to_index)

    output_file.close()


HDF5_SPECIAL_TOKENS = {".": "__DOT__", "/": "__SLASH__"}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("model_desc", help="Name of model")
    parser.add_argument(
        "input_corpus", help="Text file path with one sentence per line"
    )
    parser.add_argument(
        "output_file",
        help="Output file path where extracted representations will be stored",
    )
    parser.add_argument(
        "--aggregation",
        help="first, last or average aggregation for word representation in the case of subword segmentation",
        default="last",
    )
    parser.add_argument(
        "--output-type",
        choices=["hdf5", "json"],
        default="json",
        help="Output format of the extracted representations",
    )
    parser.add_argument("--disable_cuda", action="store_true")
    parser.add_argument("--ignore_embeddings", action="store_true")
    parser.add_argument(
        "--random_weights",
        action="store_true",
        help="generate representations from randomly initialized model",
    )
    args = parser.parse_args()

    assert args.aggregation in [
        "average",
        "first",
        "last",
    ], "Invalid aggregation option, please specify first, average or last."

    if not args.disable_cuda and torch.cuda.is_available():
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    extract_representations(
        args.model_desc,
        args.input_corpus,
        args.output_file,
        device=device,
        aggregation=args.aggregation,
        output_type=args.output_type,
        random_weights=args.random_weights,
        ignore_embeddings=args.ignore_embeddings,
    )


if __name__ == "__main__":
    main()



"""Module for layer and neuron level linear-probe based analysis.

This module contains functions to train, evaluate and use a linear probe for
both layer-wise and neuron-wise analysis.

.. seealso::
        `Dalvi, Fahim, et al. "What is one grain of sand in the desert? analyzing individual neurons in deep nlp models." Proceedings of the AAAI Conference on Artificial Intelligence. Vol. 33. No. 01. 2019. <https://ojs.aaai.org/index.php/AAAI/article/view/4592>`_
"""
import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable

from . import metrics
from . import utils

class LinearProbe(nn.Module):
    """Torch model for linear probe"""
    def __init__(self, input_size, num_classes):
        """Initialize a linear model"""
        super(LinearProbe, self).__init__()
        self.linear = nn.Linear(input_size, num_classes)

    def forward(self, x):
        """Run a forward pass on the model"""
        out = self.linear(x)
        return out

################################# Regularizers #################################
def l1_penalty(var):
    """
    L1/Lasso regularization penalty

    Parameters
    ----------
    var : torch.Variable
        Torch variable representing the weight matrix over which the penalty
        should be computed

    Returns
    -------
    penalty : torch.Variable
        Torch variable containing the penalty as a single floating point value

    """
    return torch.abs(var).sum()

def l2_penalty(var):
    """
    L2/Ridge regularization penalty.

    Parameters
    ----------
    var : torch.Variable
        Torch variable representing the weight matrix over which the penalty
        should be computed

    Returns
    -------
    penalty : torch.Variable
        Torch variable containing the penalty as a single floating point value

    Notes
    -----
    The penalty is derived from the L2-norm, which has a square root. The exact
    optimization can also be done without the square root, but this makes no
    difference in the actual output of the optimization because of the scaling
    factor used along with the penalty.

    """
    return torch.sqrt(torch.pow(var, 2).sum())

############################ Training and Evaluation ###########################
def _train_probe(
    X_train,
    y_train,
    task_type,
    lambda_l1=0,
    lambda_l2=0,
    num_epochs=10,
    batch_size=32,
    learning_rate=0.001,
):
    """
    Internal helper method to train a linear probe.

    This method is used internally for both classification and regression based
    tasks in order to train probes for them. A logistic regression model
    is trained with Cross Entropy loss for classification tasks and a linear
    regression model is trained with MSE loss for regression tasks. The
    optimizer used is Adam with default ``torch.optim`` hyperparameters.

    Parameters
    ----------
    X_train : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``
    y_train : numpy.ndarray
        Numpy Vector of size [``NUM_TOKENS``] with class labels for each input
        token. For classification, 0-indexed class labels for each input token
        are expected. For regression, a real value per input token is expected.
        Usually the output of ``interpretation.utils.create_tensors``.
    task_type : str
        Either "classification" or "regression", indicate the kind of task that
        is being probed.
    lambda_l1 : float, optional
        L1 Penalty weight in the overall loss. Defaults to 0, i.e. no L1
        regularization
    lambda_l2 : float, optional
        L2 Penalty weight in the overall loss. Defaults to 0, i.e. no L2
        regularization
    num_epochs : int, optional
        Number of epochs to train the linear model for. Defaults to 10
    batch_size : int, optional
        Batch size for the input to the linear model. Defaults to 32
    learning_rate : float, optional
        Learning rate for optimizing the linear model. 

    Returns
    -------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe for the given task.

    """
    progressbar = utils.get_progress_bar()
    print("Training %s probe" % (task_type))
    # Check if we can use GPU's for training
    use_gpu = torch.cuda.is_available()

    if lambda_l1 is None or lambda_l2 is None:
        print("Please provide regularizer weights")
        return

    print("Creating model...")
    if task_type == "classification":
        num_classes = len(set(y_train))
        assert (
            num_classes > 1
        ), "Classification problem must have more than one target class"
    else:
        num_classes = 1
    print("Number of training instances:", X_train.shape[0])
    if task_type == "classification":
        print("Number of classes:", num_classes)

    probe = LinearProbe(X_train.shape[1], num_classes)
    if use_gpu:
        probe = probe.cuda()

    if task_type == "classification":
        criterion = nn.CrossEntropyLoss()
    elif task_type == "regression":
        criterion = nn.MSELoss()
    else:
        assert (
            task_type == "classification" or task_type == "regression"
        ), "Invalid task type"

    optimizer = torch.optim.Adam(probe.parameters(), lr=learning_rate)

    X_tensor = torch.from_numpy(X_train)
    y_tensor = torch.from_numpy(y_train)

    for epoch in range(num_epochs):
        num_tokens = 0
        avg_loss = 0
        for inputs, labels in progressbar(
            utils.batch_generator(X_tensor, y_tensor, batch_size=batch_size),
            desc="epoch [%d/%d]" % (epoch + 1, num_epochs),
        ):
            num_tokens += inputs.shape[0]
            if use_gpu:
                inputs = inputs.cuda()
                labels = labels.cuda()
            inputs = Variable(inputs)
            labels = Variable(labels)

            # Forward + Backward + Optimize
            optimizer.zero_grad()
            outputs = probe(inputs)
            if task_type == "regression":
                outputs = outputs.squeeze()
            weights = list(probe.parameters())[0]

            loss = (
                criterion(outputs, labels)
                + lambda_l1 * l1_penalty(weights)
                + lambda_l2 * l2_penalty(weights)
            )
            loss.backward()
            optimizer.step()

            avg_loss += loss.item()

        print(
            "Epoch: [%d/%d], Loss: %.4f"
            % (epoch + 1, num_epochs, avg_loss / num_tokens)
        )

    return probe

def train_logistic_regression_probe(
    X_train,
    y_train,
    lambda_l1=0,
    lambda_l2=0,
    num_epochs=10,
    batch_size=32,
    learning_rate=0.001,
):
    """
    Train a logistic regression probe.

    This method trains a linear classifier that can be used as a probe to perform
    neuron analysis. Use this method when the task that is being probed for is a
    classification task. A logistic regression model is trained with Cross
    Entropy loss. The optimizer used is Adam with default ``torch.optim``
    package hyperparameters.

    Parameters
    ----------
    X_train : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``. ``dtype`` of the
        matrix must be ``np.float32``
    y_train : numpy.ndarray
        Numpy Vector with 0-indexed class labels for each input token. The size
        of the vector must be [``NUM_TOKENS``].  Usually the output of
        ``interpretation.utils.create_tensors``. Assumes that class labels are
        continuous from ``0`` to ``NUM_CLASSES-1``. ``dtype`` of the
        matrix must be ``np.int``
    lambda_l1 : float, optional
        L1 Penalty weight in the overall loss. Defaults to 0, i.e. no L1
        regularization
    lambda_l2 : float, optional
        L2 Penalty weight in the overall loss. Defaults to 0, i.e. no L2
        regularization
    num_epochs : int, optional
        Number of epochs to train the linear model for. Defaults to 10
    batch_size : int, optional
        Batch size for the input to the linear model. Defaults to 32
    learning_rate : float, optional
        Learning rate for optimizing the linear model. 

    Returns
    -------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe for the given task.

    """
    return _train_probe(
        X_train,
        y_train,
        task_type="classification",
        lambda_l1=lambda_l1,
        lambda_l2=lambda_l2,
        num_epochs=num_epochs,
        batch_size=batch_size,
        learning_rate=learning_rate,
    )


def train_linear_regression_probe(
    X_train,
    y_train,
    lambda_l1=None,
    lambda_l2=None,
    num_epochs=10,
    batch_size=32,
    learning_rate=0.001,
):
    """
    Train a linear regression probe.

    This method trains a linear classifier that can be used as a probe to perform
    neuron analysis. Use this method when the task that is being probed for is a
    regression task. A linear regression model is trained with MSE loss. The
    optimizer used is Adam with default ``torch.optim`` package hyperparameters.

    Parameters
    ----------
    X_train : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``. ``dtype`` of the
        matrix must be ``np.float32``
    y_train : numpy.ndarray
        Numpy Vector with real-valued labels for each input token. The size
        of the vector must be [``NUM_TOKENS``].  Usually the output of
        ``interpretation.utils.create_tensors``. ``dtype`` of the
        matrix must be ``np.float32``
    lambda_l1 : float, optional
        L1 Penalty weight in the overall loss. Defaults to 0, i.e. no L1
        regularization
    lambda_l2 : float, optional
        L2 Penalty weight in the overall loss. Defaults to 0, i.e. no L2
        regularization
    num_epochs : int, optional
        Number of epochs to train the linear model for. Defaults to 10
    batch_size : int, optional
        Batch size for the input to the linear model. Defaults to 32
    learning_rate : float, optional
        Learning rate for optimizing the linear model. 

    Returns
    -------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe for the given task.

    """
    return _train_probe(
        X_train,
        y_train,
        model_type="regression",
        lambda_l1=lambda_l1,
        lambda_l2=lambda_l2,
        num_epochs=num_epochs,
        batch_size=batch_size,
        learning_rate=learning_rate,
    )


def evaluate_probe(
    probe,
    X,
    y,
    idx_to_class=None,
    return_predictions=False,
    source_tokens=None,
    batch_size=32,
    metric="accuracy",
):
    """
    Evaluates a trained probe.

    This method evaluates a trained probe on the given data, and supports
    several standard metrics.

    Parameters
    ----------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe model
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``. ``dtype`` of the
        matrix must be ``np.float32``
    y : numpy.ndarray
        Numpy Vector of size [``NUM_TOKENS``] with class labels for each input
        token. For classification, 0-indexed class labels for each input token
        are expected. For regression, a real value per input token is expected.
        Usually the output of ``interpretation.utils.create_tensors``
    idx_to_class : dict, optional
        Class index to name mapping. Usually returned by
        ``interpretation.utils.create_tensors``. If this mapping is provided,
        per-class metrics are also computed. Defaults to None.
    return_predictions : bool, optional
        If set to True, actual predictions are also returned along with scores
        for further use. Defaults to False.
    source_tokens : list of lists, optional
        List of all sentences, where each is a list of the tokens in that
        sentence. Usually returned by ``data.loader.load_data``. If provided and
        ``return_predictions`` is True, each prediction will be paired with its
        original token. Defaults to None.
    batch_size : int, optional
        Batch size for the input to the model. Defaults to 32
    metrics : str, optional
        Metric to use for evaluation scores. For supported metrics see
        ``interpretation.metrics``

    Returns
    -------
    scores : dict
        The overall score on the given data with the key ``__OVERALL__``. If
        ``idx_to_class`` mapping is provided, additional keys representing each
        class and their associated scores are also part of the dictionary.
    predictions : list of 3-tuples, optional
        If ``return_predictions`` is set to True, this list will contain a
        3-tuple for every input sample, representing 
        ``(source_token, predicted_class, was_predicted_correctly)``

    """
    progressbar = utils.get_progress_bar()

    # Check if we can use GPU's for evaluation
    use_gpu = torch.cuda.is_available()

    if use_gpu:
        probe = probe.cuda()

    # Test the Model
    y_pred = []

    def source_generator():
        for s in source_tokens:
            for t in s:
                yield t

    src_words = source_generator()

    if return_predictions:
        predictions = []
        src_word = -1

    for inputs, labels in progressbar(
        utils.batch_generator(
            torch.from_numpy(X), torch.from_numpy(y), batch_size=batch_size
        ),
        desc="Evaluating",
    ):
        if use_gpu:
            inputs = inputs.cuda()
            labels = labels.cuda()
        inputs = Variable(inputs)
        labels = Variable(labels)

        outputs = probe(inputs)

        if outputs.data.shape[1] == 1:
            # Regression
            predicted = outputs.data
        else:
            # Classification
            _, predicted = torch.max(outputs.data, 1)
        predicted = predicted.cpu().numpy()

        for i in range(0, len(predicted)):
            idx = predicted[i]
            if idx_to_class:
                key = idx_to_class[idx]
            else:
                key = idx

            y_pred.append(predicted[i])

            if return_predictions:
                if source_tokens:
                    src_word = next(src_words)
                else:
                    src_word = src_word + 1
                predictions.append((src_word, key, labels[i].item() == idx))

    y_pred = np.array(y_pred)

    result = metrics.compute_score(y_pred, y, metric)

    print("Score (%s) of the probe: %0.2f" % (metric, result))

    class_scores = {}
    class_scores["__OVERALL__"] = result

    if idx_to_class:
        for i in idx_to_class:
            class_name = idx_to_class[i]
            class_instances_idx = np.where(y == i)[0]
            y_pred_filtered = y_pred[class_instances_idx]
            y_filtered = y[class_instances_idx]
            total = y_filtered.shape
            if total == 0:
                class_scores[class_name] = 0
            else:
                class_scores[class_name] = metrics.compute_score(
                    y_pred_filtered, y_filtered, metric
                )

    if return_predictions:
        return class_scores, predictions
    return class_scores

############################### Neuron Selection ###############################
def get_top_neurons(probe, percentage, class_to_idx):
    """
    Get top neurons from a trained probe.

    This method returns the set of all top neurons based on the given percentage.
    It also returns top neurons per class. All neurons (sorted by weight in
    ascending order) that account for ``percentage`` of the total weight mass
    are returned. See the given reference for the compcomplete selection algorithm
    description.

    .. seealso::
        `Dalvi, Fahim, et al. "What is one grain of sand in the desert? analyzing individual neurons in deep nlp models." Proceedings of the AAAI Conference on Artificial Intelligence. Vol. 33. No. 01. 2019. <https://ojs.aaai.org/index.php/AAAI/article/view/4592>`_

    .. note::
        Absolute weight values are used for selection, instead of raw signed
        values

    Parameters
    ----------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe model
    percentage : float
        Real number between 0 and 1, with 0 representing no weight mass and 1
        representing the entire weight mass, i.e. all neurons.
    class_to_idx : dict
        Class to class index mapping. Usually returned by
        ``interpretation.utils.create_tensors``.

    Returns
    -------
    overall_top_neurons : numpy.ndarray
        Numpy array with all top neurons
    top_neurons : dict
        Dictionary with top neurons for every class, with the class name as the
        key and ``numpy.ndarray`` of top neurons (for that class) as the value.

    Notes
    -----
    - One can expect distributed tasks to have more top neurons than focused tasks
    - One can also expect complex tasks to have more top neurons than simpler tasks

    """
    weights = list(probe.parameters())[0].data.cpu()
    weights = np.abs(weights.numpy())
    top_neurons = {}
    for c in class_to_idx:
        total_mass = np.sum(weights[class_to_idx[c], :])
        sort_idx = np.argsort(weights[class_to_idx[c], :])[::-1]
        cum_sums = np.cumsum(weights[class_to_idx[c], sort_idx])
        unselected_neurons = np.where(cum_sums >= total_mass * percentage)[0]
        if unselected_neurons.shape[0] == 0:
            selected_neurons = np.arange(cum_sums.shape[0])
        else:
            selected_neurons = np.arange(unselected_neurons[0] + 1)
        top_neurons[c] = sort_idx[selected_neurons]

    top_neurons_union = set()
    for k in top_neurons:
        for t_n in top_neurons[k]:
            top_neurons_union.add(t_n)

    return np.array(list(top_neurons_union)), top_neurons


def get_top_neurons_hard_threshold(probe, fraction, class_to_idx):
    """
    Get top neurons from a trained probe based on the maximum weight.

    This method returns the set of all top neurons based on the given threshold.
    All neurons that have a weight above ``threshold * max_weight`` are
    considered as top neurons. It also returns top neurons per class. 

    .. note::
        Absolute weight values are used for selection, instead of raw signed
        values

    Parameters
    ----------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe model
    fraction : float
        Fraction of maximum weight per class to use for selection
    class_to_idx : dict
        Class to class index mapping. Usually returned by
        ``interpretation.utils.create_tensors``.

    Returns
    -------
    overall_top_neurons : numpy.ndarray
        Numpy array with all top neurons
    top_neurons : dict
        Dictionary with top neurons for every class, with the class name as the
        key and ``numpy.ndarray`` of top neurons (for that class) as the value.

    """
    weights = list(probe.parameters())[0].data.cpu()
    weights = np.abs(weights.numpy())
    top_neurons = {}
    for c in class_to_idx:
        top_neurons[c] = np.where(
            weights[class_to_idx[c], :]
            > np.max(weights[class_to_idx[c], :]) / fraction
        )[0]

    top_neurons_union = set()
    for k in top_neurons:
        for t_n in top_neurons[k]:
            top_neurons_union.add(t_n)

    return np.array(list(top_neurons_union)), top_neurons


def get_bottom_neurons(probe, percentage, class_to_idx):
    """
    Get bottom neurons from a trained probe.

    Analogous to ``interpretation.linear_probe.get_top_neurons``. This method
    returns the set of all bottom neurons based on the given percentage.
    It also returns bottom neurons per class. All neurons (sorted by weight
    in ascending order) that account for ``percentage`` of the total weight mass
    are returned. See the given reference for the complete selection algorithm
    description.

    .. seealso::
        `Dalvi, Fahim, et al. "What is one grain of sand in the desert? analyzing individual neurons in deep nlp models." Proceedings of the AAAI Conference on Artificial Intelligence. Vol. 33. No. 01. 2019. <https://ojs.aaai.org/index.php/AAAI/article/view/4592>`_

    .. note::
        Absolute weight values are used for selection, instead of raw signed
        values

    Parameters
    ----------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe model
    percentage : float
        Real number between 0 and 1, with 0 representing no weight mass and 1
        representing the entire weight mass, i.e. all neurons.
    class_to_idx : dict
        Class to class index mapping. Usually returned by
        ``interpretation.utils.create_tensors``.

    Returns
    -------
    overall_bottom_neurons : numpy.ndarray
        Numpy array with all bottom neurons
    bottom_neurons : dict
        Dictionary with bottom neurons for every class, with the class name as the
        key and ``numpy.ndarray`` of bottom neurons (for that class) as the value.

    """
    weights = list(probe.parameters())[0].data.cpu()
    weights = np.abs(weights.numpy())

    bottom_neurons = {}
    for c in class_to_idx:
        total_mass = np.sum(weights[class_to_idx[c], :])
        sort_idx = np.argsort(weights[class_to_idx[c], :])
        cum_sums = np.cumsum(weights[class_to_idx[c], sort_idx])
        unselected_neurons = np.where(cum_sums >= total_mass * percentage)[0]
        if unselected_neurons.shape[0] == 0:
            selected_neurons = np.arange(cum_sums.shape[0])
        else:
            selected_neurons = np.arange(unselected_neurons[0] + 1)
        bottom_neurons[c] = sort_idx[selected_neurons]

    bottom_neurons_union = set()
    for k in bottom_neurons:
        for t_n in bottom_neurons[k]:
            bottom_neurons_union.add(t_n)

    return np.array(list(bottom_neurons_union)), bottom_neurons


def get_random_neurons(probe, probability):
    """
    Get random neurons from a trained probe.

    This method returns a random set of neurons based on the probability. Each
    neuron is either discarded or included based on a uniform random variable's
    value (included if its less than probability, discarded otherwise)

    Parameters
    ----------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe model
    probability : float
        Real number between 0 and 1, with 0 representing no selection and 1
        representing selection of all neurons.

    Returns
    -------
    random_neurons : numpy.ndarray
        Numpy array with random neurons

    """
    weights = list(probe.parameters())[0].data.cpu()
    weights = np.abs(weights.numpy())

    mask = np.random.random((weights.shape[1],))
    idx = np.where(mask <= probability)[0]

    return idx


def get_neuron_ordering(probe, class_to_idx, search_stride=100):
    """
    Get global ordering of neurons from a trained probe.

    This method returns the global ordering of neurons in a model based on
    the given probe's weight values. Top neurons are computed at increasing
    percentages of the weight mass and then accumulated in-order. See given
    reference for a complete description of the selection algorithm.

    For example, if the neuron list at 1% weight mass is [#2, #52, #134], and 
    at 2% weight mass is [#2, #4, #52, #123, #130, #134, #567], the returned
    ordering will be [#2, #52, #134, #4, #123, #130, #567]. 
    Within each percentage, the ordering of neurons is arbitrary. In this case,
    the importance of #2, #52 and #134 is not necessarily in that order.
    The cutoffs between each percentage selection are also returned. Increasing
    the ``search_stride`` will decrease the distance between each cutoff, making
    the overall ordering more accurate.

    .. seealso::
        `Dalvi, Fahim, et al. "What is one grain of sand in the desert? analyzing individual neurons in deep nlp models." Proceedings of the AAAI Conference on Artificial Intelligence. Vol. 33. No. 01. 2019. <https://ojs.aaai.org/index.php/AAAI/article/view/4592>`_

    .. note::
        Absolute weight values are used for selection, instead of raw signed
        values


    Parameters
    ----------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe model
    class_to_idx : dict
        Class to class index mapping. Usually returned by
        ``interpretation.utils.create_tensors``.
    search_stride : int, optional
        Defines how many pieces the percent weight mass selection is divided
        into. Higher leads to more a accurate ordering. Defaults to 100.

    Returns
    -------
    global_neuron_ordering : numpy.ndarray
        Numpy array of size ``NUM_NEURONS`` with neurons in decreasing order
        of importance.
    cutoffs : list
        Indices where each percentage selection begins. All neurons between two
        cutoff values are arbitrarily ordered.

    """
    progressbar = utils.get_progress_bar()

    neuron_orderings = [
        get_top_neurons(probe, p / search_stride, class_to_idx)[0]
        for p in progressbar(range(search_stride + 1))
    ]

    considered_neurons = set()
    ordering = []
    cutoffs = []
    for local_ordering in neuron_orderings:
        local_ordering = list(local_ordering)
        new_neurons = set(local_ordering).difference(considered_neurons)
        ordering = ordering + list(new_neurons)
        considered_neurons = considered_neurons.union(new_neurons)

        cutoffs.append(len(ordering))

    return ordering, cutoffs


def get_neuron_ordering_granular(
    probe, class_to_idx, granularity=50, search_stride=100
):
    """
    Get global ordering of neurons from a trained probe.

    This method is an alternative to
    ``interpretation.linear_probe.get_neuron_ordering``. It works very similarly
    to that method, except that instead of adding the neurons from each
    percentage selection, neurons are added in chunks of ``granularity``
    neurons.

    .. seealso::
        `Dalvi, Fahim, et al. "What is one grain of sand in the desert? analyzing individual neurons in deep nlp models." Proceedings of the AAAI Conference on Artificial Intelligence. Vol. 33. No. 01. 2019. <https://ojs.aaai.org/index.php/AAAI/article/view/4592>`_

    .. note::
        Absolute weight values are used for selection, instead of raw signed
        values


    Parameters
    ----------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe model
    class_to_idx : dict
        Class to class index mapping. Usually returned by
        ``interpretation.utils.create_tensors``.
    granularity : int, optional
        Approximate number of neurons in each chunk of selection. Defaults to
        50.
    search_stride : int, optional
        Defines how many pieces the percent weight mass selection is divided
        into. Higher leads to more a accurate ordering. Defaults to 100.

    Returns
    -------
    global_neuron_ordering : numpy.ndarray
        Numpy array of size ``NUM_NEURONS`` with neurons in decreasing order
        of importance.
    cutoffs : list
        Indices where each chunk of selection begins. Each chunk will contain
        approximately ``granularity`` neurons. All neurons between two
        cutoff values (i.e. a chunk) are arbitrarily ordered.

    """
    progressbar = utils.get_progress_bar()

    weights = list(probe.parameters())[0].data.cpu()
    num_neurons = weights.numpy().shape[1]
    neuron_orderings = [
        get_top_neurons(probe, p / search_stride, class_to_idx)[0]
        for p in progressbar(range(search_stride + 1))
    ]

    sliding_idx = 0
    considered_neurons = set()
    ordering = []
    cutoffs = []
    for i in range(0, num_neurons + 1, granularity):
        while len(neuron_orderings[sliding_idx]) < i:
            sliding_idx = sliding_idx + 1
        new_neurons = set(neuron_orderings[sliding_idx]).difference(considered_neurons)
        if len(new_neurons) != 0:
            ordering = ordering + list(new_neurons)
            considered_neurons = considered_neurons.union(new_neurons)

            cutoffs.append(len(ordering))

    return ordering, cutoffs

# Returns num_bottom_neurons bottom neurons from the global ordering
def get_fixed_number_of_bottom_neurons(probe, num_bottom_neurons, class_to_idx):
    """
    Get global bottom neurons.

    This method returns a fixed number of bottoms neurons from the global
    ordering computed using ``interpretation.linear_probe.get_neuron_ordering``.

    .. note::
        Absolute weight values are used for selection, instead of raw signed
        values


    Parameters
    ----------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe model
    num_bottom_neurons : int
        Number of bottom neurons for selection
    class_to_idx : dict
        Class to class index mapping. Usually returned by
        ``interpretation.utils.create_tensors``.

    Returns
    -------
    global_bottom_neurons : numpy.ndarray
        Numpy array of size ``num_bottom_neurons`` with bottom neurons using the
        global ordering

    """
    ordering, _ = get_neuron_ordering(probe, class_to_idx)

    return ordering[-num_bottom_neurons:]


import math
import numpy as np

from imblearn.under_sampling import RandomUnderSampler

def isnotebook():
    """
    Utility function to detect if the code being run is within a jupyter
    notebook. Useful to change progress indicators for example.

    Returns
    -------
    isnotebook : bool
        True if the function is being called inside a notebook, False otherwise.
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter

def get_progress_bar():
    """
    Utility function to get a progress bar depending on the environment the code
    is running in. A normal text-based progress bar is returned in normal
    shells, and a notebook widget-based progress bar is returned in jupyter
    notebooks.

    Returns
    -------
    progressbar : function
        The appropriate progressbar from the tqdm library.

    """
    if isnotebook():
        from tqdm import tqdm_notebook as progressbar
    else:
        from tqdm import tqdm as progressbar

    return progressbar

def batch_generator(X, y, batch_size=32):
    """
    Generator function to generate batches of data for training/evaluation.

    This function takes two tensors representing the activations and labels
    respectively, and yields batches of parallel data. The last batch may
    contain fewer than ``batch_size`` elements.

    Parameters
    ----------
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually the
        output of ``interpretation.utils.create_tensors``
    y : numpy.ndarray
        Numpy Vector of size [``NUM_TOKENS``] with class labels for each input
        token. For classification, 0-indexed class labels for each input token
        are expected. For regression, a real value per input token is expected.
        Usually the output of ``interpretation.utils.create_tensors``
    batch_size : int, optional
        Number of samples to return in each call. Defaults to 32.

    Yields
    ------
    X_batch : numpy.ndarray
        Numpy Matrix of size [``batch_size`` x ``NUM_NEURONS``]. The final batch
        may have fewer elements than the requested ``batch_size``
    y_batch : numpy.ndarray
        Numpy Vector of size [``batch_size``]. The final batch may have fewer
        elements than the requested ``batch_size``
    """
    start_idx = 0
    while start_idx < X.shape[0]:
        yield X[start_idx : start_idx + batch_size], y[
            start_idx : start_idx + batch_size
        ]
        start_idx = start_idx + batch_size

def tok2idx(tokens):
    """
    Utility function to generate unique indices for a set of tokens.

    Parameters
    ----------
    tokens : list of lists
        List of sentences, where each sentence is a list of tokens. Usually
        returned from ``data.loader.load_data``

    Returns
    -------
    tok2idx_mapping : dict
        A dictionary with tokens as keys and a unique index for each token as
        values
    """
    uniq_tokens = set().union(*tokens)
    return {p: idx for idx, p in enumerate(uniq_tokens)}


def idx2tok(srcidx):
    """
    Utility function to an inverse mapping from a ``tok2idx`` mapping.

    Parameters
    ----------
    tok2idx_mapping : dict
        Token to index mapping, usually the output for
        ``interpretation.utils.tok2idx``.

    Returns
    -------
    idx2tok : dict
        A dictionary with unique indices as keys and their associated tokens as
        values
    """
    return {v: k for k, v in srcidx.items()}


def count_target_words(tokens):
    """
    Utility function to count the total number of tokens in a dataset.

    Parameters
    ----------
    tokens : list of lists
        List of sentences, where each sentence is a list of tokens. Usually
        returned from ``data.loader.load_data``

    Returns
    -------
    count : int
        Total number of tokens in the given ``tokens`` structure
    """
    return sum([len(t) for t in tokens["target"]])


def create_tensors(
    tokens, activations, task_specific_tag, mappings=None, task_type="classification"
):
    """
    Method to pre-process loaded datasets into tensors that can be used to train
    probes and perform analyis on. The input tokens are represented as list of
    sentences, where each sentence is a list of tokens. Each token also has
    an associated label. All tokens from all sentences are flattened into one
    dimension in the returned tensors. The returned tensors will thus have
    ``total_num_tokens`` rows.

    Parameters
    ----------
    tokens : list of lists
        List of sentences, where each sentence is a list of tokens. Usually
        returned from ``data.loader.load_data``
    activations : list of numpy.ndarray
        List of *sentence representations*, where each *sentence representation*
        is a numpy matrix of shape
        ``[num tokens in sentence x concatenated representation size]``. Usually
        retured from ``data.loader.load_activations``
    task_specific_tag : str
        Label to assign tokens with unseen labels. This is particularly useful
        if some labels are never seen during train, but are present in the dev
        or test set. This is usually set to the majority class in the task.
    mappings : list of dicts
        List of four python dicts: ``label2idx``, ``idx2label``, ``src2idx`` and
        ``idx2src`` for classification tasks. List of two dicts  ``src2idx`` and
        ``idx2src`` for regression tasks. Each dict represents either the
        mapping from class labels to indices and source tokens to indices or
        vice versa. Usually returned from a previous call to ``create_tensors``.
    task_type : str
        Either "classification" or "regression", indicate the kind of task that
        is being probed.

    Returns
    -------
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]
    y : numpy.ndarray
        Numpy vector of size [``NUM_TOKENS``]
    mappings : list of dicts
        List of four python dicts: ``label2idx``, ``idx2label``, ``src2idx`` and
        ``idx2src`` for classification tasks. List of two dicts  ``src2idx`` and
        ``idx2src`` for regression tasks. Each dict represents either the
        mapping from class labels to indices and source tokens to indices or
        vice versa.

    Notes
    -----
    - ``mappings`` should be created exactly once, and should be reused for subsequent calls
    - For example, ``mappings`` can be created on train data, and the passed during the call for dev and test data.

    """
    assert (
        task_type == "classification" or task_type == "regression"
    ), "Invalid model type"
    num_tokens = count_target_words(tokens)
    print("Number of tokens: ", num_tokens)

    num_neurons = activations[0].shape[1]

    source_tokens = tokens["source"]
    target_tokens = tokens["target"]

    ####### creating pos and source to index and reverse
    if mappings is not None:
        if task_type == "classification":
            label2idx, idx2label, src2idx, idx2src = mappings
        else:
            src2idx, idx2src = mappings
    else:
        if task_type == "classification":
            label2idx = tok2idx(target_tokens)
            idx2label = idx2tok(label2idx)
        src2idx = tok2idx(source_tokens)
        idx2src = idx2tok(src2idx)

    print("length of source dictionary: ", len(src2idx))
    if task_type == "classification":
        print("length of target dictionary: ", len(label2idx))

    X = np.zeros((num_tokens, num_neurons), dtype=np.float32)
    if task_type=="classification":
        y = np.zeros((num_tokens,), dtype=np.int)
    else:
        y = np.zeros((num_tokens,), dtype=np.float32)

    example_set = set()

    idx = 0
    for instance_idx, instance in enumerate(target_tokens):
        for token_idx, _ in enumerate(instance):
            if idx < num_tokens:
                X[idx] = activations[instance_idx][token_idx, :]

            example_set.add(source_tokens[instance_idx][token_idx])
            if task_type == "classification":
                if (
                    mappings is not None
                    and target_tokens[instance_idx][token_idx] not in label2idx
                ):
                    y[idx] = label2idx[task_specific_tag]
                else:
                    y[idx] = label2idx[target_tokens[instance_idx][token_idx]]
            elif task_type == "regression":
                y[idx] = float(target_tokens[instance_idx][token_idx])

            idx += 1

    print(idx)
    print("Total instances: %d" % (num_tokens))
    print(list(example_set)[:20])

    if task_type == "classification":
        return X, y, (label2idx, idx2label, src2idx, idx2src)
    return X, y, (src2idx, idx2src)


################################## Statictics ##################################
def print_overall_stats(all_results):
    """
    Method to pretty print overall results.

    .. warning::
        This method was primarily written to process results from internal
        scripts and pipelines.

    Parameters
    ----------
    all_results : dict
        Dictionary containing the probe, overall scores, scores from selected
        neurons, neuron ordering and neuron selections at various percentages

    """
    probe = all_results["probe"]
    weights = list(probe.parameters())[0].data.cpu()
    num_neurons = weights.numpy().shape[1]
    print(
        "Overall accuracy: %0.02f%%"
        % (100 * all_results["original_accs"]["__OVERALL__"])
    )

    print("")
    print("Global results")
    print("10% Neurons")
    print(
        "\tKeep Top accuracy: %0.02f%%"
        % (100 * all_results["global_results"]["10%"]["keep_top_accs"]["__OVERALL__"])
    )
    print(
        "\tKeep Random accuracy: %0.02f%%"
        % (
            100
            * all_results["global_results"]["10%"]["keep_random_accs"]["__OVERALL__"]
        )
    )
    print(
        "\tKeep Bottom accuracy: %0.02f%%"
        % (
            100
            * all_results["global_results"]["10%"]["keep_bottom_accs"]["__OVERALL__"]
        )
    )
    print("15% Neurons")
    print(
        "\tKeep Top accuracy: %0.02f%%"
        % (100 * all_results["global_results"]["15%"]["keep_top_accs"]["__OVERALL__"])
    )
    print(
        "\tKeep Random accuracy: %0.02f%%"
        % (
            100
            * all_results["global_results"]["15%"]["keep_random_accs"]["__OVERALL__"]
        )
    )
    print(
        "\tKeep Bottom accuracy: %0.02f%%"
        % (
            100
            * all_results["global_results"]["15%"]["keep_bottom_accs"]["__OVERALL__"]
        )
    )
    print("20% Neurons")
    print(
        "\tKeep Top accuracy: %0.02f%%"
        % (100 * all_results["global_results"]["20%"]["keep_top_accs"]["__OVERALL__"])
    )
    print(
        "\tKeep Random accuracy: %0.02f%%"
        % (
            100
            * all_results["global_results"]["20%"]["keep_random_accs"]["__OVERALL__"]
        )
    )
    print(
        "\tKeep Bottom accuracy: %0.02f%%"
        % (
            100
            * all_results["global_results"]["20%"]["keep_bottom_accs"]["__OVERALL__"]
        )
    )
    print("")
    print("Full order of neurons:")
    print(all_results["global_results"]["ordering"])

    print("--------------------")
    print("")
    print("Local results")
    for idx, percentage in enumerate(all_results["local_results"]["percentages"]):
        print("Weight Mass percentage: %d%%" % (percentage * 100))
        _, top_neurons, top_neurons_per_tag = all_results["local_results"][
            "local_top_neurons"
        ][idx]
        print(
            "Percentage of all neurons: %0.0f%%"
            % (100 * len(top_neurons) / num_neurons)
        )
        print("Top Neurons:", sorted(top_neurons))
        print("")
        print("Top neurons per tag:")
        for tag in top_neurons_per_tag:
            print("\t" + tag + ":", sorted(top_neurons_per_tag[tag]))
            print("")


def print_machine_stats(all_results):
    """
    Method to print overall results in tsv format.

    .. warning::
        This method was primarily written to process results from internal
        scripts and pipelines.

    Parameters
    ----------
    all_results : dict
        Dictionary containing the probe, overall scores, scores from selected
        neurons, neuron ordering and neuron selections at various percentages

    """

    probe = all_results["probe"]
    weights = list(probe.parameters())[0].data.cpu()
    num_neurons = weights.numpy().shape[1]
    print("Filtering out:")
    print(
        "%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%s"
        % (
            100 * all_results["original_accs"]["__OVERALL__"],
            100 * all_results["global_results"]["10%"]["keep_top_accs"]["__OVERALL__"],
            100
            * all_results["global_results"]["10%"]["keep_random_accs"]["__OVERALL__"],
            100
            * all_results["global_results"]["10%"]["keep_bottom_accs"]["__OVERALL__"],
            100 * all_results["global_results"]["15%"]["keep_top_accs"]["__OVERALL__"],
            100
            * all_results["global_results"]["15%"]["keep_random_accs"]["__OVERALL__"],
            100
            * all_results["global_results"]["15%"]["keep_bottom_accs"]["__OVERALL__"],
            100 * all_results["global_results"]["20%"]["keep_top_accs"]["__OVERALL__"],
            100
            * all_results["global_results"]["20%"]["keep_random_accs"]["__OVERALL__"],
            100
            * all_results["global_results"]["20%"]["keep_bottom_accs"]["__OVERALL__"],
            str(all_results["global_results"]["ordering"][:300]),
        )
    )
    print("\nZero out:")
    print(
        "%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f\t%0.2f"
        % (
            100 * all_results["original_accs"]["__OVERALL__"],
            100
            * all_results["global_results"]["10%"]["zero_out_top_accs"]["__OVERALL__"],
            100
            * all_results["global_results"]["10%"]["zero_out_random_accs"][
                "__OVERALL__"
            ],
            100
            * all_results["global_results"]["10%"]["zero_out_bottom_accs"][
                "__OVERALL__"
            ],
            100
            * all_results["global_results"]["15%"]["zero_out_top_accs"]["__OVERALL__"],
            100
            * all_results["global_results"]["15%"]["zero_out_random_accs"][
                "__OVERALL__"
            ],
            100
            * all_results["global_results"]["15%"]["zero_out_bottom_accs"][
                "__OVERALL__"
            ],
            100
            * all_results["global_results"]["20%"]["zero_out_top_accs"]["__OVERALL__"],
            100
            * all_results["global_results"]["20%"]["zero_out_random_accs"][
                "__OVERALL__"
            ],
            100
            * all_results["global_results"]["20%"]["zero_out_bottom_accs"][
                "__OVERALL__"
            ],
        )
    )

    for idx, percentage in enumerate(all_results["local_results"]["percentages"]):
        print("\nLocal %d%%:" % (percentage * 100))
        top_neurons = all_results["local_results"]["local_top_neurons"][idx][1]
        top_neurons_per_tag = all_results["local_results"]["local_top_neurons"][idx][2]
        top_neurons_per_tag_list = {k: list(v) for k, v in top_neurons_per_tag.items()}
        print(
            "%0.2f%%\t%s\t%s"
            % (
                100 * len(top_neurons) / num_neurons,
                str(sorted(top_neurons)),
                str(top_neurons_per_tag_list),
            )
        )


################################ Data Balancing ################################
def balance_binary_class_data(X, y):
    """
    Method to balance binary class data.

    The majority class is under-sampled randomly to match the minority class in
    it's size.

    Parameters
    ----------
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually
        returned from ``interpretation.utils.create_tensors``
    y : numpy.ndarray
        Numpy vector of size [``NUM_TOKENS``]. Usually returned from
        ``interpretation.utils.create_tensors``

    Returns
    -------
    X_balanced : numpy.ndarray
        Numpy matrix of size [``NUM_BALANCED_TOKENS`` x ``NUM_NEURONS``]
    y_balanced : numpy.ndarray
        Numpy vector of size [``NUM_BALANCED_TOKENS``]

    """
    rus = RandomUnderSampler()
    X_res, y_res = rus.fit_resample(X, y)

    return X_res, y_res


def balance_multi_class_data(X, y, num_required_instances=None):
    """
    Method to balance multi class data.

    All classes are under-sampled randomly to match the minority class in
    their size. If ``num_required_instances`` is provided, all classes are
    sampled proportionally so that the total number of selected examples is
    approximately ``num_required_instances`` (because of rounding proportions).

    Parameters
    ----------
    X : numpy.ndarray
        Numpy Matrix of size [``NUM_TOKENS`` x ``NUM_NEURONS``]. Usually
        returned from ``interpretation.utils.create_tensors``
    y : numpy.ndarray
        Numpy vector of size [``NUM_TOKENS``]. Usually returned from
        ``interpretation.utils.create_tensors``
    num_required_instances : int, optional
        Total number of required instances. All classes are sampled
        proportionally.

    Returns
    -------
    X_balanced : numpy.ndarray
        Numpy matrix of size [``NUM_BALANCED_TOKENS`` x ``NUM_NEURONS``]
    y_balanced : numpy.ndarray
        Numpy vector of size [``NUM_BALANCED_TOKENS``]

    """
    if num_required_instances:
        total = y.shape[0]
        unique, counts = np.unique(y, return_counts=True)
        class_counts = dict(zip(unique, counts))
        num_instances_per_class = {
            key: math.ceil(count / total * num_required_instances)
            for key, count in class_counts.items()
        }
        print(num_instances_per_class)
        rus = RandomUnderSampler(sampling_strategy=num_instances_per_class)
    else:
        rus = RandomUnderSampler()

    X_res, y_res = rus.fit_resample(X, y)

    return X_res, y_res

def load_probe(probe_path):
    """
    Loads a probe and its associated mappings from probe_path

    .. warning::
        This method is currently not implemented.

    Parameters
    ----------
    probe_path : str
        Path to a pkl object saved by interpretation.utils.save_probe

    Returns
    -------
    probe : interpretation.linear_probe.LinearProbe
        Trained probe model
    mappings : list of dicts
        List of four python dicts: ``label2idx``, ``idx2label``, ``src2idx`` and
        ``idx2src`` for classification tasks. List of two dicts  ``src2idx`` and
        ``idx2src`` for regression tasks. Each dict represents either the
        mapping from class labels to indices and source tokens to indices or
        vice versa.

    """
    pass

def save_probe(probe_path, probe, mappings):
    """
    Saves a model and its associated mappings as a pkl object at probe_path

    .. warning::
        This method is currently not implemented.

    Parameters
    ----------
    probe_path : str
        Path to save a pkl object
    probe : interpretation.linear_probe.LinearProbe
        Trained probe model
    mappings : list of dicts
        List of four python dicts: ``label2idx``, ``idx2label``, ``src2idx`` and
        ``idx2src`` for classification tasks. List of two dicts  ``src2idx`` and
        ``idx2src`` for regression tasks. Each dict represents either the
        mapping from class labels to indices and source tokens to indices or
        vice versa.

    """
    pass